

import numpy as np

# ToFu-specific
import tofu.geom._GG as _GG



"""
def calc_directproblem(pts=None, r=None,
                       Ves=None, LStruct=None, res=None,
                       demiss=None, ani=False, axisym=True,
                       approx=True, out_coefonly=False,
                       Forbid=True, verb=True, Test=True):

    # Preformat
    if verb:
        s0 = "    Direct problem"
    nt = pts.shape[1]
    if Ves is not None:
        VPoly, VIn, Lim, VType = Ves.Poly, Ves.geom['VIn'], Ves.Lim, Ves.Type
        if LStruct is not None:
            LSPoly = [ss.Poly for ss in LStruct]
            LSVIn = [ss.geom['VIn'] for ss in LStruct]
            LSLim = [ss.Lim for ss in LStruct]
        else:
            LSPoly, LSVIn, LSLim = None, None, None
    else:
        VPoly, VIn, Lim, VType = None, None, None, None

    DV = [None, None, DPhi]
    pts, dV, ind, dVr = Ves.get_sampleV(res, DV=DV, dVMode='abs', ind=None, Out='(X,Y,Z)'):
    nR =

    dout = {''}
    if demiss is None or demiss=={}:
        sang = _GG.Dust_calc_SolidAngle(pos, r, pts,
                                        approx=approx, out_coefonly=out_coefonly,
                                        VType=VType, VPoly=VPoly, VIn=VIn,
                                        VLim=Lim, LSPoly=LSPoly, LSLim=LSLim,
                                        LSVIn=LSVIn, Forbid=Forbid, Test=Test):
        dout['sang':out[0]]
    elif not ani and axisym:
        dout = _GG.calc_SolidAngleView_emissIsoAxisym()
    else:
        dout = _GG.calc_SolidAngleView_emiss()
    else:
        raise(Exception, "Not coded yet !")




    for ii in range(0,nt):
        if verb:
            print(s0+": step {0}/{1}".format(ii,nt))
        for jj in range(0,nR):
            pp =
            vis =
            sa =



    lpolyCross, lpolyHor, gridCross, gridHor = out[:4]
    saCross, saHor, volIn, saIn = out[4:8]
    contribCross, contribHor, powIn
"""



# -*- coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# Also if needed: retab
'''
    TEST equimap
'''
from __future__ import (unicode_literals, absolute_import,  \
                        print_function, division)
import matplotlib.pyplot as plt
import numpy as np
import os
import sys
import time
#import warnings

if __name__ == '__main__':
    #print('path 1 =', sys.path)
    sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
    #print('path 2 =', sys.path)

    # Local modules
    import imas
    import equimap
    #import imas_west
    #import pywed as pw

    shot    = 53221
    tol_val = 1E-10

    # For 2D plots
    interp_points = 60

    # FIRST POINT B_NORM
    # ------------------
    time_in = np.linspace(36, 37, 10)
    Phi_in  = np.linspace(0, 2*np.pi/18, 100)
    R_in    = np.full(Phi_in.shape, 3)
    Z_in    = np.zeros(R_in.shape)

    # Read equilibrium data
    idd = imas.ids(shot, 0)
    idd.open_env('imas_public', 'west', '3')
    idd.equilibrium.get()
    out = idd.equilibrium

    equiDict = {}

    # Declaration of arrays 2d plots
    equi_grid = idd.equilibrium.grids_ggd[0].grid[0]
    NbrPoints = len(equi_grid.space[0].objects_per_dimension[0].object)
    equiDict['r'] = np.full(NbrPoints, np.nan)
    equiDict['z'] = np.full(NbrPoints, np.nan)
    for ii in range(NbrPoints):
        equiDict['r'][ii] = equi_grid.space[0].objects_per_dimension[0]. \
                            object[ii].geometry[0]
        equiDict['z'][ii] = equi_grid.space[0].objects_per_dimension[0]. \
                            object[ii].geometry[1]

    # For 2D plots
    R_all = np.linspace(np.min(equiDict['r']), np.max(equiDict['r']), interp_points)
    Z_all = np.linspace(np.min(equiDict['z']), np.max(equiDict['z']), interp_points)

    R_all_tot = np.repeat(R_all, interp_points)
    Z_all_tot = np.tile(Z_all, interp_points)

    Rr = R_all_tot.reshape((interp_points, interp_points))
    Zr = Z_all_tot.reshape((interp_points, interp_points))

    # CALL EQUIMAP
    start = time.time()
    oute = equimap.get(shot, time=time_in, \
            R=R_in, Phi=Phi_in, Z=Z_in, \
            quantity='b_field_norm')
    end = time.time()
    print()
    print('time in equimap.get b_norm =', end - start)
    print()
    print('oute.shape b_norm =', oute.shape)

    # CALL EQUIMAP
    start = time.time()
    oute_noR = equimap.get(shot, time=time_in, \
                 R=R_in, Phi=Phi_in, Z=Z_in, \
                 quantity='b_field_norm', no_ripple=True)
    end = time.time()
    print()
    print('time in equimap.get b_norm no Ripple =', end - start)
    print()
    print('oute.shape b_norm no ripple =', oute_noR.shape)

    print()
    print('Mean value B_norm ripple =', np.mean(oute[int(0.5*oute.shape[0]), :]))
    print('Mean value B_norm NO ripple =', \
                np.mean(oute_noR[int(0.5*oute_noR.shape[0]), :]))
    diff_mean_val = np.mean(oute[int(0.5*oute.shape[0]), :]) \
                  - np.mean(oute_noR[int(0.5*oute_noR.shape[0]), :])
    print('Diff mean values =', diff_mean_val)
    percent_diff = np.abs(100*diff_mean_val \
                 / np.mean(oute[int(0.5*oute.shape[0]), :]))
    print('Percent diff mean values =', percent_diff)

    # CHECK
    # -----
    if (np.abs(percent_diff - 0.011052598088) > tol_val):
        print()
        print('ERROR: Higher than tolerance percent difference ' \
                     + str(np.abs(percent_diff - 0.011052598088)))
        print()
        #raise RuntimeError
    # FOR:
    # shot    = 53221
    # time_in = np.linspace(36, 37, 10)
    # Phi_in  = np.linspace(0, 2*np.pi/18, 100)
    # R_in    = np.full(Phi_in.shape, 3)
    # Z_in    = np.zeros(R_in.shape)
    # RESULTS:
    # Mean value B_norm ripple = 3.05593472975
    # Mean value B_norm NO ripple = 3.05627248994
    # Diff mean values = -0.000337760183512
    # Percent diff mean values = 0.011052598088
    print()

    # PLOTS
    plt.figure()
    plt.plot(time_in, oute[:, -1], label='B_norm at R={0}, Phi=Z=0'.format(R_in[-1]))
    plt.plot(time_in, oute_noR[:, -1], label='B_norm no ripple at R={0}, Phi=Z=0'.format(R_in[-1]))
    plt.legend()
    plt.xlabel('Time [s]')
    plt.ylabel('B_norm [T]')
    plt.figure()
    plt.plot(Phi_in, oute[int(0.5*oute.shape[0]), :], \
             label='B_norm at t={0:.2f}, R={1}, Z=0'.format( \
                   time_in[int(0.5*oute.shape[0])], R_in[-1]))
    plt.plot(Phi_in, oute_noR[int(0.5*oute.shape[0]), :], \
             label='B_norm no ripple at t={0:.2f}, R={1}, Z=0'.format( \
                   time_in[int(0.5*oute.shape[0])], R_in[-1]))
    plt.legend()
    plt.xlabel('Phi [rad]')
    plt.ylabel('B_norm [T]')

    # SECOND POSITION B_NORM
    # ----------------------
    t_ignitron = []
    t_ignitron.append(32)
    print()
    print('t_igni =', t_ignitron[0])
    print()
    time_in = np.linspace(t_ignitron[0], 38, 10)
    Phi_in = np.linspace(0, 2*np.pi/18, 100)
    R_in   = np.full(Phi_in.shape, 2.43)
    Z_in   = np.full(Phi_in.shape, 0.57)

    # CALL EQUIMAP
    start = time.time()
    oute = equimap.get(shot, time=time_in, \
            R=R_in, Phi=Phi_in, Z=Z_in, \
            quantity='b_field_norm')
    end = time.time()
    print()
    print('time in equimap.get 2 b_norm =', end - start)
    print()
    print('oute.shape 2 b_norm =', oute.shape)

    # CALL EQUIMAP
    start = time.time()
    oute_noR = equimap.get(shot, time=time_in, \
                 R=R_in, Phi=Phi_in, Z=Z_in, \
                 quantity='b_field_norm', no_ripple=True)
    end = time.time()
    print()
    print('time in equimap.get 2 b_norm no ripple =', end - start)
    print()
    print('oute.shape 2 b_norm no ripple =', oute_noR.shape)

    # PLOTS
    plt.figure()
    plt.plot(time_in, oute[:, -1], \
      label='B_norm at R={0}, Phi={1:.2f}, Z={2}'.format( \
            R_in[-1], Phi_in[-1], Z_in[-1]))
    plt.plot(time_in, oute_noR[:, -1], \
      label='B_norm no ripple at R={0}, Phi={1:.2f}, Z={2}'.format( \
            R_in[-1], Phi_in[-1], Z_in[-1]))
    plt.legend()
    plt.xlabel('Time [s]')
    plt.ylabel('B_norm [T]')
    plt.figure()
    plt.plot(Phi_in, oute[int(0.5*oute.shape[0]), :], \
             label='B_norm at t={0:.2f}, R={1}, Z={2}'.format( \
                   time_in[int(0.5*oute.shape[0])], R_in[-1], Z_in[-1]))
    plt.plot(Phi_in, oute_noR[int(0.5*oute.shape[0]), :], \
             label='B_norm no ripple at t={0:.2f}, R={1}, Z={2}'.format( \
                   time_in[int(0.5*oute.shape[0])], R_in[-1], Z_in[-1]))
    plt.legend()
    plt.xlabel('Phi [rad]')
    plt.ylabel('B_norm [T]')

    # B_NORM 2D
    # ---------
    # CALL EQUIMAP
    start = time.time()
    outa = equimap.get(shot, time=time_in, \
            R=R_all_tot, Phi=np.zeros(R_all_tot.shape), Z=Z_all_tot, \
            quantity='b_field_norm')
    end = time.time()
    print()
    print('time in equimap.get b_norm 2D =', end - start)
    print()
    outar = outa[int(0.5*outa.shape[0])].reshape((interp_points, interp_points))
    plt.figure()
    plt.contourf(Rr, Zr, outar)
    plt.colorbar()
    arg_time = np.argmin(np.abs(out.time - time_in[int(0.5*outa.shape[0])]))
    plt.plot(np.squeeze(out.time_slice[arg_time].boundary.outline.r), \
             np.squeeze(out.time_slice[arg_time].boundary.outline.z), \
             linewidth=2, color='red')
    plt.plot(out.time_slice[arg_time].global_quantities.magnetic_axis.r, \
             out.time_slice[arg_time].global_quantities.magnetic_axis.z, \
             marker='+', color='red', markersize=20)
    plt.xlabel('R [m]')
    plt.ylabel('Z [m]')
    plt.title('B_norm t={0:.2f}'.format(time_in[int(0.5*outa.shape[0])]))

    # B_R TEST
    # --------
    Phi_in = np.linspace(0, 2*np.pi/18, 100)
    R_in   = np.full(Phi_in.shape, 3)
    Z_in   = np.full(Phi_in.shape, 0)

    # CALL EQUIMAP
    start = time.time()
    oute = equimap.get(shot, time=time_in, \
            R=R_in, Phi=Phi_in, Z=Z_in, \
            quantity='b_field_r')
    end = time.time()
    print()
    print('time in equimap.get br =', end - start)
    print()
    print('oute.shape br =', oute.shape)

    # CALL EQUIMAP
    start = time.time()
    oute_noR = equimap.get(shot, time=time_in, \
                 R=R_in, Phi=Phi_in, Z=Z_in, \
                 quantity='b_field_r', no_ripple=True)
    end = time.time()
    print()
    print('time in equimap.get br no ripple =', end - start)
    print()
    print('oute.shape br no ripple =', oute_noR.shape)

    # PLOTS
    plt.figure()
    plt.plot(time_in, oute[:, -1], \
      label='B_r at R={0}, Phi={1:.2f}, Z={2}'.format( \
            R_in[-1], Phi_in[-1], Z_in[-1]))
    plt.plot(time_in, oute_noR[:, -1], \
      label='B_r no ripple at R={0}, Phi={1:.2f}, Z={2}'.format( \
            R_in[-1], Phi_in[-1], Z_in[-1]))
    plt.legend()
    plt.xlabel('Time [s]')
    plt.ylabel('B_r [T]')

    plt.figure()
    plt.plot(Phi_in, oute[int(0.5*oute.shape[0]), :], \
             label='B_r at t={0:.2f}, R={1}, Z={2}'.format( \
                   time_in[int(0.5*oute.shape[0])], R_in[-1], Z_in[-1]))
    plt.plot(Phi_in, oute_noR[int(0.5*oute.shape[0]), :], \
             label='B_r no ripple at t={0:.2f}, R={1}, Z={2}'.format( \
                   time_in[int(0.5*oute.shape[0])], R_in[-1], Z_in[-1]))
    plt.legend()
    plt.xlabel('Phi [rad]')
    plt.ylabel('B_r [T]')

    # CALL EQUIMAP
    start = time.time()
    outa = equimap.get(shot, time=time_in, \
            R=R_all_tot, Phi=np.zeros(R_all_tot.shape), Z=Z_all_tot, \
            quantity='b_field_r')
    end = time.time()
    print()
    print('time in equimap.get br 2D =', end - start)
    print()
    outar = outa[int(0.5*outa.shape[0])].reshape((interp_points, interp_points))
    plt.figure()
    plt.contourf(Rr, Zr, outar)
    plt.colorbar()
    arg_time = np.argmin(np.abs(out.time - time_in[int(0.5*outa.shape[0])]))
    plt.plot(np.squeeze(out.time_slice[arg_time].boundary.outline.r), \
             np.squeeze(out.time_slice[arg_time].boundary.outline.z), \
             linewidth=2, color='red')
    plt.plot(out.time_slice[arg_time].global_quantities.magnetic_axis.r, \
             out.time_slice[arg_time].global_quantities.magnetic_axis.z, \
             marker='+', color='red', markersize=20)
    plt.xlabel('R [m]')
    plt.ylabel('Z [m]')
    plt.title('B_r t={0:.2f}'.format(time_in[int(0.5*outa.shape[0])]))

    # B_Z TEST
    # --------
    Phi_in = np.linspace(0, 2*np.pi/18, 100)
    R_in   = np.full(Phi_in.shape, 3)
    Z_in   = np.full(Phi_in.shape, 0.2)

    # CALL EQUIMAP
    start = time.time()
    oute = equimap.get(shot, time=time_in, \
            R=R_in, Phi=Phi_in, Z=Z_in, \
            quantity='b_field_z')
    end = time.time()
    print()
    print('time in equimap.get bz =', end - start)
    print()
    print('oute.shape bz =', oute.shape)

    # CALL EQUIMAP
    start = time.time()
    oute_noR = equimap.get(shot, time=time_in, \
                 R=R_in, Phi=Phi_in, Z=Z_in, \
                 quantity='b_field_z', no_ripple=True)
    end = time.time()
    print()
    print('time in equimap.get bz no ripple =', end - start)
    print()
    print('oute.shape bz no ripple =', oute_noR.shape)

    # PLOTS
    plt.figure()
    plt.plot(time_in, oute[:, -1], \
      label='B_z at R={0}, Phi={1:.2f}, Z={2}'.format( \
            R_in[-1], Phi_in[-1], Z_in[-1]))
    plt.plot(time_in, oute_noR[:, -1], \
      label='B_z no ripple at R={0}, Phi={1:.2f}, Z={2}'.format( \
            R_in[-1], Phi_in[-1], Z_in[-1]))
    plt.legend()
    plt.xlabel('Time [s]')
    plt.ylabel('B_z [T]')

    plt.figure()
    plt.plot(Phi_in, oute[int(0.5*oute.shape[0]), :], \
             label='B_z at t={0:.2f}, R={1}, Z={2}'.format( \
                   time_in[int(0.5*oute.shape[0])], R_in[-1], Z_in[-1]))
    plt.plot(Phi_in, oute_noR[int(0.5*oute.shape[0]), :], \
             label='B_z no ripple at t={0:.2f}, R={1}, Z={2}'.format( \
                   time_in[int(0.5*oute.shape[0])], R_in[-1], Z_in[-1]))
    plt.legend()
    plt.xlabel('Phi [rad]')
    plt.ylabel('B_z [T]')

    # CALL EQUIMAP
    start = time.time()
    outa = equimap.get(shot, time=time_in, \
            R=R_all_tot, Phi=np.zeros(R_all_tot.shape), Z=Z_all_tot, \
            quantity='b_field_z')
    end = time.time()
    print()
    print('time in equimap.get bz 2D =', end - start)
    print()
    outar = outa[int(0.5*outa.shape[0])].reshape((interp_points, interp_points))
    plt.figure()
    plt.contourf(Rr, Zr, outar)
    plt.colorbar()
    arg_time = np.argmin(np.abs(out.time - time_in[int(0.5*outa.shape[0])]))
    plt.plot(np.squeeze(out.time_slice[arg_time].boundary.outline.r), \
             np.squeeze(out.time_slice[arg_time].boundary.outline.z), \
             linewidth=2, color='red')
    plt.plot(out.time_slice[arg_time].global_quantities.magnetic_axis.r, \
             out.time_slice[arg_time].global_quantities.magnetic_axis.z, \
             marker='+', color='red', markersize=20)
    plt.xlabel('R [m]')
    plt.ylabel('Z [m]')
    plt.title('B_z t={0:.2f}'.format(time_in[int(0.5*outa.shape[0])]))

    # B_TOR TEST
    # ----------
    Phi_in = np.linspace(0, 2*np.pi/18, 100)
    R_in   = np.full(Phi_in.shape, 3)
    Z_in   = np.full(Phi_in.shape, 0)

    # CALL EQUIMAP
    start = time.time()
    oute = equimap.get(shot, time=time_in, \
            R=R_in, Phi=Phi_in, Z=Z_in, \
            quantity='b_field_tor')
    end = time.time()
    print()
    print('time in equimap.get btor =', end - start)
    print()
    print('oute.shape btor =', oute.shape)

    # CALL EQUIMAP
    start = time.time()
    oute_noR = equimap.get(shot, time=time_in, \
                 R=R_in, Phi=Phi_in, Z=Z_in, \
                 quantity='b_field_tor', no_ripple=True)
    end = time.time()
    print()
    print('time in equimap.get btor no ripple =', end - start)
    print()
    print('oute.shape btor no ripple =', oute_noR.shape)

    # PLOTS
    plt.figure()
    plt.plot(time_in, oute[:, -1], \
      label='B_tor at R={0}, Phi={1:.2f}, Z={2}'.format( \
            R_in[-1], Phi_in[-1], Z_in[-1]))
    plt.plot(time_in, oute_noR[:, -1], \
      label='B_tor no ripple at R={0}, Phi={1:.2f}, Z={2}'.format( \
            R_in[-1], Phi_in[-1], Z_in[-1]))
    plt.legend()
    plt.xlabel('Time [s]')
    plt.ylabel('B_tor [T]')

    plt.figure()
    plt.plot(Phi_in, oute[int(0.5*oute.shape[0]), :], \
             label='B_tor at t={0:.2f}, R={1}, Z={2}'.format( \
                   time_in[int(0.5*oute.shape[0])], R_in[-1], Z_in[-1]))
    plt.plot(Phi_in, oute_noR[int(0.5*oute.shape[0]), :], \
             label='B_tor no ripple at t={0:.2f}, R={1}, Z={2}'.format( \
                   time_in[int(0.5*oute.shape[0])], R_in[-1], Z_in[-1]))
    plt.legend()
    plt.xlabel('Phi [rad]')
    plt.ylabel('B_tor [T]')

    # CALL EQUIMAP
    start = time.time()
    outa = equimap.get(shot, time=time_in, \
            R=R_all_tot, Phi=np.zeros(R_all_tot.shape), Z=Z_all_tot, \
            quantity='b_field_tor')
    end = time.time()
    print()
    print('time in equimap.get btor 2D =', end - start)
    print()
    outar = outa[int(0.5*outa.shape[0])].reshape((interp_points, interp_points))
    plt.figure()
    plt.contourf(Rr, Zr, outar)
    plt.colorbar()
    arg_time = np.argmin(np.abs(out.time - time_in[int(0.5*outa.shape[0])]))
    plt.plot(np.squeeze(out.time_slice[arg_time].boundary.outline.r), \
             np.squeeze(out.time_slice[arg_time].boundary.outline.z), \
             linewidth=2, color='red')
    plt.plot(out.time_slice[arg_time].global_quantities.magnetic_axis.r, \
             out.time_slice[arg_time].global_quantities.magnetic_axis.z, \
             marker='+', color='red', markersize=20)
    plt.xlabel('R [m]')
    plt.ylabel('Z [m]')
    plt.title('B_tor t={0:.2f}'.format(time_in[int(0.5*outa.shape[0])]))

    plt.show()



# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
#
################################################################################
# Utility functions related to the distance between a LOS and a VPoly:
#    - Compute distance between LOS and a circle (unit and vectorial)
#    - Test if a LOS and a circle are close (unit and vectorial) to a epsilon
#    - Compute distance between LOS and VPoly (unit and vectorial)
#    - Test if a LOS and a VPoly are close (unit and vectorial) to a epsilon
#    - Compute which LOS is closer to a VPoly
#    - Compute which VPoly is closer to a LOS
################################################################################

# ==============================================================================
# == DISTANCE CIRCLE - LOS
# ==============================================================================
cdef void dist_los_circle_core(const double[3] direct,
                                      const double[3] origin,
                                      const double radius, const double circ_z,
                                      double norm_dir,
                                      double[2] result) nogil

cdef void comp_dist_los_circle_vec_core(int num_los, int num_cir,
                                               double* los_directions,
                                               double* los_origins,
                                               double* circle_radius,
                                               double* circle_z,
                                               double* norm_dir_tab,
                                               double[::1] res_k,
                                               double[::1] res_dist,
                                               int num_threads) nogil

# ==============================================================================
# == TEST CLOSENESS CIRCLE - LOS
# ==============================================================================
cdef bint is_close_los_circle_core(const double[3] direct,
                                          const double[3] origin,
                                          double radius, double circ_z,
                                          double norm_dir, double eps) nogil

cdef void is_close_los_circle_vec_core(int num_los, int num_cir,
                                       double eps,
                                       double* los_directions,
                                       double* los_origins,
                                       double* circle_radius,
                                       double* circle_z,
                                       double* norm_dir_tab,
                                       int[::1] res, int num_threads) nogil

# ==============================================================================
# == DISTANCE BETWEEN LOS AND EXT-POLY
# ==============================================================================
cdef void comp_dist_los_vpoly_vec_core(int num_poly, int nlos,
                                              double* ray_orig,
                                              double* ray_vdir,
                                              double[:,:,::1] ves_poly,
                                              double eps_uz,
                                              double eps_a,
                                              double eps_vz,
                                              double eps_b,
                                              double eps_plane,
                                              int ves_type,
                                              int algo_type,
                                              double* res_k,
                                              double* res_dist,
                                              double disc_step,
                                              int num_threads) nogil

cdef void simple_dist_los_vpoly_core(const double[3] ray_orig,
                                            const double[3] ray_vdir,
                                            const double* lpolyx,
                                            const double* lpolyy,
                                            const int nvert,
                                            const double upscaDp,
                                            const double upar2,
                                            const double dpar2,
                                            const double invuz,
                                            const double crit2,
                                            const double eps_uz,
                                            const double eps_vz,
                                            const double eps_a,
                                            const double eps_b,
                                            double* res_final) nogil

# ==============================================================================
# == ARE LOS AND EXT-POLY CLOSE
# ==============================================================================
cdef void is_close_los_vpoly_vec_core(int num_poly, int nlos,
                                             double* ray_orig,
                                             double* ray_vdir,
                                             double[:,:,::1] ves_poly,
                                             double eps_uz,
                                             double eps_a,
                                             double eps_vz,
                                             double eps_b,
                                             double eps_plane,
                                             double epsilon,
                                             int[::1] are_close,
                                             int num_threads) nogil

# ==============================================================================
# == WHICH LOS/VPOLY IS CLOSER
# ==============================================================================
cdef void which_los_closer_vpoly_vec_core(int num_poly, int nlos,
                                                 double* ray_orig,
                                                 double* ray_vdir,
                                                 double[:,:,::1] ves_poly,
                                                 double eps_uz,
                                                 double eps_a,
                                                 double eps_vz,
                                                 double eps_b,
                                                 double eps_plane,
                                                 int[::1] ind_close_tab,
                                                 int num_threads) nogil

cdef void which_vpoly_closer_los_vec_core(int num_poly, int nlos,
                                                 double* ray_orig,
                                                 double* ray_vdir,
                                                 double[:,:,::1] ves_poly,
                                                 double eps_uz,
                                                 double eps_a,
                                                 double eps_vz,
                                                 double eps_b,
                                                 double eps_plane,
                                                 int[::1] ind_close_tab,
                                                 int num_threads) nogil



"""
Creating your first Geometry
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

This is a tutorial that shows you how to create a simple geometry.
"""

import numpy as np
import matplotlib.pyplot as plt
import tofu.geom as tfg


###############################################################################
# Creating an empty Vessel
# ------------------------
#
# If a vessel object does not exist yet, you have to create one (otherwise you
# can just load an existing one).
# A vessel object is basically defined by a 2D simple polygon
# (i.e.: non self-intersecting), that is then expanded linearly or toroidally
# depending on the desired configuration.
# This polygon limits the volume available for the plasma, where the emissivity
# can be non-zero. It is typically defined by the inner wall in a tokamak.
#
# Let's define the polygon limiting the vessel as a circle with a divertor-like
# shape at the bottom:


# Define the center, radius and lower limit
R0, Z0, rad, ZL = 2.0, 0.0, 1.0, -0.85
# Define the key points in the divertor region below ZL
Div_R, Div_Z = [R0 - 0.2, R0, R0 + 0.2], [-1.2, -0.9, -1.2]
# Find the angles corresponding to ZL and span the rest
thet1 = np.arcsin((ZL - Z0) / rad)
thet2 = np.pi - thet1
thet = np.linspace(thet1, thet2, 100)
# Assemble the polygon
poly_R = np.append(R0 + rad * np.cos(thet), Div_R)
poly_Z = np.append(Z0 + rad * np.sin(thet), Div_Z)
# Plot for checking
plt.figure(facecolor="w", figsize=(6, 6))
plt.plot(poly_R, poly_Z)
plt.axis("equal")

###############################################################################
# Notice that the polygon does not have to be closed, ToFu will anyway check
# that and close it automatically if necessary.
# Now let's feed this 2D polygon to the appropriate ToFu class and specify that
# it should be a toroidal type (if linear type 'Lin' is chosen, the length
# should be specified by the 'Lim' keyword argument).
# **tofu** also asks for a name to be associated to this instance, and an
# experiment ('Exp') and a shot number (useful when the same experiment changes
# geometry in time).

# Create a toroidal Ves instance with name 'MyFirstVessel', associated to
# the experiment 'Misc' (for 'Miscellaneous') and shot number 0
ves = tfg.Ves(
    Name="MyFirstVessel",
    Poly=[poly_R, poly_Z],
    Type="Tor",
    Exp="Misc",
    shot=0
)

###############################################################################
# Now the vessel instance is created. It provides you several key attributes
# and methods (see :class:`~tofu.geom` for details).
# Among them the Id attribute is itself a class instance that contains all
# useful information about this vessel instance for identification, saving...
# In particular, that's where the name, the default saving path, the Type, the
# experiment, the shot number... are all stored.
# A default name for saving was also created that automatically includes not
# only the name you gave but also the module from which this instance was
# created (tofu.geom or tfg), the type of object, the experiment, the shot
# number...
# This recommended default pattern is useful for quick identification of saved
# object, it is advised not to modify it.

print(ves.Id.SaveName)

###############################################################################
# Now, we can simply visualise the created vessel by using the dedicated method
# (keyword argument 'Elt' specifies the elements of the instance we want to
# plot, typically one letter corresponds to one element, here we just want the
# polygon):

# Plot the polygon by default in two projections (cross-section and horizontal)
# and return the list of axes
lax = ves.plot(element="P")


###############################################################################
# The created vessel instance, plotted in cross-section and horizontal
# projections
# Since the vessel is an important object (it defines where the plasma lives),
# all the other ToFu objects rely on it. It is thus important that you save
# it so that it can be used by other ToFu objects when necessary.

ves.save(path="./")

###############################################################################
# This method will save the instance as a numpy compressed file (.npz), using
# the path and file name found in ves.Id.SavePath and ves.Id.SaveName. While
# it is highly recommended to stick to the default value for the SaveName,
# but you can easily modify the saving path if you want by specifying it using
# keyword argument Path.

###############################################################################
# Adding structural elements
# ---------------------------
#
# Like for a vessel, a structural element is mostly defined by a 2D polygon.
# If a vessel instance is provided, the type of the structural element
# (toroidal or linear) is automatically the same as the type of the vessel,
# otherwise the type must be specified.

# A configuration, short for geometrical configuration is a set of vessel,
# and structural elements.

# Define two polygons, one that does not enclose the vessel and one that does
thet = np.linspace(0.0, 2.0 * np.pi, 100)
poly1 = [[2.5, 3.5, 3.5, 2.5], [0.0, 0.0, 0.5, 0.5]]  # a rectangle
poly2 = [R0 + 0.5 * np.cos(thet), -1.0 + 0.5 * np.sin(thet)]  # a circle
poly3 = [[0.8, 1.3, 1.3, 0.8], [-0.5, -0.5, 0.5, 0.5]]  # another rectangle
# Create the structural elements with the appropriate ToFu class, specifying
# the experiment and a shot number for keeping track of changes
s1 = tfg.PFC(Name="S1",
             Poly=poly1,
             Exp="Misc",
             shot=0)
# now we create a structure that is not continuous along phi
# but is only defined within certain limits
s2 = tfg.PFC(
    Name="S2",
    Poly=poly2,
    Exp="Misc",
    shot=0,
    Lim=[[0.0, np.pi], [np.pi / 2.0, np.pi * 3.0 / 2.0]],
)
# and another one, now defined as repetitions centered a position `pos`
# and with a certain `extent`
# here we wanted a structure uniformly repeated 5 times along phi
s3 = tfg.PFC(
    Name="S3",
    Poly=poly3,
    # we dont take the last element of the list as it is 2.*pi = 0
    pos=np.linspace(0.0, 2.0 * np.pi, 6)[:-1],
    # 5 repetitions + 5 empty spaces = 10 subdivision of 2. pi
    extent=np.pi * 2.0 / 10.0,
    Exp="Misc",
    shot=0,
)
# Creating a configuration with vessel and structures
config = tfg.Config(Name="test",
                    Exp="Misc",
                    lStruct=[ves, s1, s2, s3])
config.set_colors_random()  # to see different colors
config.plot()
config.save()
# sphinx_gallery_thumbnail_number = 3



import os
import shutil
import subprocess
import matplotlib.pyplot as plt
from tofu import __version__
from pathlib import Path

_HERE = os.path.abspath(os.path.dirname(__file__))
_TFROOT = Path(_HERE).parent.parent.parent

keyVers = 'Vers'


#######################################################
#
#     Setup and Teardown
#
#######################################################

def setup_module(module, here=_HERE):
    print("")   # this is to get a newline after the dots
    lf = os.listdir(here)
    lf = [f for f in lf
          if all([s in f for s in ['.npz']])]
    lF = []
    for f in lf:
        ff = f.split('_')
        v = [fff[len(keyVers):] for fff in ff
             if fff[:len(keyVers)] == keyVers]
        msg = f + "\n    " + str(ff) + "\n    " + str(v)
        assert len(v) == 1, msg
        v = v[0]
        if '.npz' in v:
            v = v[:v.index('.npz')]
        # print(v, __version__)
        if v != __version__:
            lF.append(f)
    if len(lF) > 0:
        print("Removing the following previous test files:")
        for f in lF:
            os.remove(os.path.join(here, f))
        # print("setup_module before anything in this file")


def teardown_module(module, here=_HERE):
    # os.remove(VesTor.Id.SavePath + VesTor.Id.SaveName + '.npz')
    # os.remove(VesLin.Id.SavePath + VesLin.Id.SaveName + '.npz')
    # print("teardown_module after everything in this file")
    # print("") # this is to get a newline
    lf = os.listdir(here)
    lf = [f for f in lf
          if all([s in f for s in ['.npz']])]
    lF = []
    for f in lf:
        ff = f.split('_')
        v = [fff[len(keyVers):] for fff in ff
             if fff[:len(keyVers)] == keyVers]
        msg = f + "\n    "+str(ff) + "\n    " + str(v)
        assert len(v) == 1, msg
        v = v[0]
        if '.npz' in v:
            v = v[:v.index('.npz')]
        # print(v, __version__)
        if v == __version__:
            lF.append(f)
    if len(lF) > 0:
        print("Removing the following test files:")
        for f in lF:
            os.remove(os.path.join(here, f))


#######################################################
#
#   Struct subclasses
#
#######################################################

def get_list_toturials(path=_HERE):
    ltut = os.listdir(path)
    ltut = [tt[:-3] for tt in ltut
            if (all([ss in tt for ss in ['tuto', '.py']])
                and not any([ss in tt for ss in ['__init__', '.swp']]))]
    return ltut


class Test00_tuto(object):

    @classmethod
    def setup_class(cls, here=_HERE, pathtuto=_HERE, root=_TFROOT):
        # print("")
        # print("---- "+cls.__name__)
        # ii = 0
        # for tuto in get_list_toturials(path=pathtuto):
        # # Create local temporary copy to make sure the
        # local version of tofu is imported
        # method = 'test{0:02.0f}_{1}'.format(ii, tuto)
        # fmethod = lambda tuto=tuto: cls.Test_tuto._test_tuto(tuto)
        # setattr(cls, method, types.MethodType(fmethod, cls))
        # # setattr(cls, method, lambda cls: cls._test_tuto(tuto))
        # ii += 1
        pass

    @classmethod
    def teardown_class(cls):
        # print("teardown_class() after any methods in this class")
        pass

    def setup_method(self):
        # print("TestUM:setup_method() before each test method")
        # self.setup_class()
        pass

    def teardown_method(self):
        # print("TestUM:teardown_method() after each test method")
        pass

    @classmethod
    def _test_tuto(cls, tuto=None, pathtuto=_HERE, root=_TFROOT):
        src = os.path.join(pathtuto, tuto + '.py')
        target = os.path.join(root, tuto + '.py')
        shutil.copyfile(src, target)
        error = None
        try:
            cmd = 'python ' + target
            out = subprocess.run(cmd, shell=True, check=False,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            ls = out.stderr.decode().split('\n')
            if any(['Error' in ss and not 'ModuleNotFoundError' in ss
                    for ss in ls]):
                error = '\n'.join(ls)
        except Exception as err:
            error = err

        if error is not None:
            msg = str(error)
            raise Exception(msg)
        plt.close('all')

        # Remove temporary files and saved files
        os.remove(target)
        stdout = out.stdout.decode().split('\n')
        lf = [stdout[ii].strip() for ii in range(len(stdout))
              if 'saved' in stdout[ii-1].lower()]
        for ii in range(len(lf)):
            os.remove(lf[ii])

    def test01_plot_basic_tutorial(self):
        self._test_tuto('tuto_plot_basic')

    def test02_plot_create_geometry(self):
        self._test_tuto('tuto_plot_create_geometry')

    def test03_plot_custom_emissivity(self):
        self._test_tuto('tuto_plot_custom_emissivity')



# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import scipy.sparse as scpsp
import datastock as ds


_LOPERATORS_INT = [
    'D1',
    'D2',
    'D3',
    'D0N1',
    'D0N2',
    'D1N2',
    'D2N2',
    'D3N2',
]


# #############################################################################
# #############################################################################
#                   Mesh2DRect - bsplines - operators
# #############################################################################


def _get_mesh2dRect_operators_check(
    deg=None,
    operator=None,
    geometry=None,
    sparse_fmt=None,
):

    # deg
    deg = ds._generic_check._check_var(
        deg, 'deg',
        types=int,
        allowed=[0, 1, 2, 3],
    )

    # operator
    operator = ds._generic_check._check_var(
        operator, 'operator',
        default='D0N1',
        types=str,
        allowed=_LOPERATORS_INT,
    )

    # geometry
    geometry = ds._generic_check._check_var(
        geometry, 'geometry',
        default='toroidal',
        types=str,
        allowed=['toroidal', 'linear'],
    )

    # sparse_fmt
    sparse_fmt = ds._generic_check._check_var(
        sparse_fmt, 'sparse_fmt',
        default='csc',
        types=str,
        allowed=['dia', 'csr', 'csc', 'lil'],
    )

    # dim
    if operator == 'D1':
        dim = 'origin / m'
    elif operator == 'D2':
        dim = 'origin / m2'
    elif operator == 'D3':
        dim = 'origin / m3'
    elif operator == 'D0N1':
        if geometry == 'linear':
            dim = 'origin x m2'
        else:
            dim = 'origin x m3/rad'
    elif operator == 'D0N2':
        if geometry == 'linear':
            dim = 'origin2 x m2'
        else:
            dim = 'origin2 x m3/rad'
    elif operator == 'D1N2':
        if geometry == 'linear':
            dim = 'origin2'
        else:
            dim = 'origin2 x m/rad'
    elif operator == 'D2N2':
        if geometry == 'linear':
            dim = 'origin2 / m2'
        else:
            dim = 'origin2 / (m2.rad)'

    return operator, geometry, sparse_fmt, dim


def get_mesh2dRect_operators(
    operator=None,
    geometry=None,
    deg=None,
    knotsx_mult=None,
    knotsy_mult=None,
    knotsx_per_bs=None,
    knotsy_per_bs=None,
    overlap=None,
    sparse_fmt=None,
    cropbs_flat=None,
    # specific to deg = 0
    cropbs=None,
):

    # ------------
    # check inputs

    operator, geometry, sparse_fmt, dim = _get_mesh2dRect_operators_check(
        deg=deg,
        operator=operator,
        geometry=geometry,
        sparse_fmt=sparse_fmt,
    )

    # ------------
    # prepare

    nx, ny = knotsx_per_bs.shape[1], knotsy_per_bs.shape[1]
    kR = np.tile(knotsx_per_bs, ny)
    kZ = np.repeat(knotsy_per_bs, nx, axis=1)
    nbs = nx*ny

    if cropbs_flat is None:
        cropbs_flat = False
    if cropbs_flat is not False:
        c0 = (
            isinstance(cropbs_flat, np.ndarray)
            and cropbs_flat.shape == (nbs,)
            and cropbs_flat.dtype == np.bool_
        )
        if not c0:
            msg = (
                f"Arg cropbs_flat must be a bool array of shape {(nbs,)}\n"
                f"Provided: {cropbs_flat.shape}"
            )
            raise Exception(msg)
        nbscrop = cropbs_flat.sum()
        shape = (nbscrop, nbscrop)
        indbs = -np.ones((nbs,), dtype=int)
        indbs[cropbs_flat] = np.arange(0, nbscrop)
    else:
        shape = (nbs, nbs)
        indbs = np.arange(0, nbs)

    if 'N2' in operator and deg >= 1:
        # get intersection indices array
        if cropbs_flat is False:
            nbtot = np.sum(overlap >= 0)
        else:
            ind = cropbs_flat[None, :] & cropbs_flat[overlap]
            nbtot = np.sum(ind)

        # prepare data and indices arrays
        if operator == 'D0N2':
            data = np.full((nbtot,), np.nan)
            row = np.zeros((nbtot,), dtype=int)
            column = np.zeros((nbtot,), dtype=int)
        elif operator == 'D1N2':
            datadR = np.full((nbtot,), np.nan)
            datadZ = np.full((nbtot,), np.nan)
            row = np.zeros((nbtot,), dtype=int)
            column = np.zeros((nbtot,), dtype=int)
        elif operator == 'D2N2':
            datad2R = np.full((nbtot,), np.nan)
            datad2Z = np.full((nbtot,), np.nan)
            datadRZ = np.full((nbtot,), np.nan)
            row = np.zeros((nbtot,), dtype=int)
            column = np.zeros((nbtot,), dtype=int)

    # ------------
    # D0 - integral

    if operator == 'D0N1':
        if deg == 0 and geometry == 'linear':

            opmat = (kR[1, :] - kR[0, :]) * (kZ[1, :] - kZ[0, :])

        elif deg == 0 and geometry == 'toroidal':

            opmat = 0.5 * (kR[1, :]**2 - kR[0, :]**2) * (kZ[1, :] - kZ[0, :])

        elif deg == 1 and geometry == 'linear':

            opmat = 0.25 * (kR[2, :] - kR[0, :]) * (kZ[2, :] - kZ[0, :])

        elif deg == 1 and geometry == 'toroidal':

            opmat = (
                0.5
                * (kR[2, :]**2 - kR[0, :]**2 + kR[1, :]*(kR[2, :]-kR[0, :]))
                * (kZ[2, :] - kZ[0, :])
            ) / 6.

        elif deg == 2:

            iZ1 = (kZ[1, :] - kZ[0, :])**2 / (3.*(kZ[2, :] - kZ[0, :]))
            iZ21 = (
                (
                    kZ[2, :]**2
                    - 2*kZ[1, :]**2
                    + kZ[1, :]*kZ[2, :]
                    + 3.*kZ[0, :]*(kZ[1, :] - kZ[2, :])
                )
                / (6.*(kZ[2, :]-kZ[0, :]))
            )
            iZ22 = (
                (
                    -2.*kZ[2, :]**2
                    + kZ[1, :]**2
                    + kZ[1, :]*kZ[2, :]
                    + 3.*kZ[3, :]*(kZ[2, :] - kZ[1, :])
                )
                / (6.*(kZ[3, :] - kZ[1, :]))
            )
            iZ3 = (kZ[3, :] - kZ[2, :])**2 / (3.*(kZ[3, :] - kZ[1, :]))

            if geometry == 'linear':
                iR1 = (kR[1, :] - kR[0, :])**2 / (3.*(kR[2, :] - kR[0, :]))
                iR21 = (
                    (
                        kR[2, :]**2
                        - 2. * kR[1, :]**2
                        + kR[1, :] * kR[2, :]
                        + 3. * kR[0, :] * (kR[1, :] - kR[2, :])
                    )
                    / (6. * (kR[2, :] - kR[0, :]))
                )
                iR22 = (
                    (
                        -2. * kR[2, :]**2
                        + kR[1, :]**2
                        + kR[1, :] * kR[2, :]
                        + 3. * kR[3, :] * (kR[2, :] - kR[1, :])
                    )
                    / (6.*(kR[3, :] - kR[1, :]))
                )
                iR3 = (kR[3, :] - kR[2, :])**2 / (3.*(kR[3, :] - kR[1, :]))

            else:
                iR1 = (
                    (
                        3.*kR[1, :]**3
                        + kR[0, :]**3
                        - 5.*kR[0, :] * kR[1, :]**2
                        + kR[0, :]**2 * kR[1, :]
                    )
                    / (12. * (kR[2, :] - kR[0, :]))
                )
                iR21 = (
                    (
                        kR[2, :]**3
                        - 3.*kR[1, :]**3
                        + kR[1, :]**2 * kR[2, :]
                        + kR[1, :] * kR[2, :]**2
                        - 2.*kR[0, :] * kR[2, :]**2
                        - 2.*kR[0, :] * kR[1, :] * kR[2, :]
                        + 4.*kR[0, :] * kR[1, :]**2
                    )
                    / (12. * (kR[2, :] - kR[0, :]))
                )
                iR22 = (
                    (
                        -3.*kR[2, :]**3
                        + kR[1, :]**3
                        + kR[1, :] * kR[2, :]**2
                        + kR[1, :]**2 * kR[2, :]
                        + 4.*kR[2, :]**2 * kR[3, :]
                        - 2.*kR[1, :]*kR[2, :]*kR[3, :]
                        - 2.*kR[1, :]**2 * kR[3, :]
                    )
                    / (12. * (kR[3, :] - kR[1, :]))
                )
                iR3 = (
                    (
                        kR[3, :]**3
                        + 3.*kR[2, :]**3
                        - 5.*kR[2, :]**2 * kR[3, :]
                        + kR[2, :]*kR[3, :]**2
                    ) / (12. * (kR[3, :] - kR[1, :]))
                )

            opmat = (iR1 + iR21 + iR22 + iR3) * (iZ1 + iZ21 + iZ22 + iZ3)

        elif deg == 3:

            msg = "Integral D0N1 not implemented for deg=3 yet!"
            raise NotImplementedError(msg)

        # crop
        if cropbs_flat is not False:
            opmat = opmat[cropbs_flat]

    # ------------
    # D1 - gradient

    elif operator == 'D1':

        # Treat separately discrete case
        if deg == 0:
            gradR, gradZ = _D1_Deg0(
                knotsx_mult=knotsx_mult,
                knotsy_mult=knotsy_mult,
                cropbs=cropbs,
                cropbs_flat=cropbs_flat,
                nx=nx,
                ny=ny,
                nbs=nbs,
            )
            opmat = (
                scpsp.csc_matrix(gradR),
                scpsp.csc_matrix(gradZ),
            )

        elif deg >= 1:
            raise NotImplementedError()

    # ------------
    # D0N2

    elif operator == 'D0N2' and deg == 0:

        iZ = kZ[1, :] - kZ[0, :]
        if geometry == 'linear':
            iR = kR[1, :] - kR[0, :]
        else:
            iR = 0.5 * (kR[1, :]**2 - kR[0, :]**2)

        if cropbs_flat is not False:
            iR = iR[cropbs_flat]
            iZ = iZ[cropbs_flat]

        opmat = scpsp.diags(
            [iR*iZ],
            [0],
            shape=None,
            format=sparse_fmt,
            dtype=float,
        )

    elif operator == 'D0N2':

        # pre-compute integrals
        if deg == 1:
            iR = _D0N2_Deg1(knotsx_mult, geometry=geometry)
            iZ = _D0N2_Deg1(knotsy_mult, geometry='linear')
        elif deg == 2:
            iR = _D0N2_Deg2(knotsx_mult, geometry=geometry)
            iZ = _D0N2_Deg2(knotsy_mult, geometry='linear')
        elif deg == 3:
            msg = "Integral D0N2 not implemented for deg=3!"
            raise NotImplementedError(msg)

        # set non-diagonal elements
        i0 = 0
        for ir in range(nx):
            for iz in range(ny):

                iflat = ir + iz*nx
                if cropbs_flat is not False and not cropbs_flat[iflat]:
                    continue

                # general case
                overlapi = overlap[:, iflat][overlap[:, iflat] > iflat]

                # diagonal element
                data[i0] = iR[0, ir] * iZ[0, iz]
                row[i0] = indbs[iflat]
                column[i0] = indbs[iflat]
                i0 += 1

                # non-diagonal elements (symmetric)
                for jflat in overlapi:

                    if cropbs_flat is not False and not cropbs_flat[jflat]:
                        continue

                    jr = jflat % nx
                    jz = jflat // nx

                    # store (i, j) and (j, i) (symmetric matrix)
                    if jr >= ir:
                        iiR = iR[jr - ir, ir]
                    else:
                        iiR = iR[abs(jr - ir), jr]
                    if jz >= iz:
                        iiZ = iZ[jz - iz, iz]
                    else:
                        iiZ = iZ[abs(jz - iz), jz]
                    data[i0:i0+2] = iiR * iiZ
                    row[i0:i0+2] = (indbs[iflat], indbs[jflat])
                    column[i0:i0+2] = (indbs[jflat], indbs[iflat])
                    i0 += 2

        assert i0 == nbtot
        opmat = scpsp.csc_matrix((data, (row, column)), shape=shape)

    # ------------
    # D1N2

    elif operator == 'D1N2':

        # Treat separately discrete case
        if deg == 0:
            gradR, gradZ = _D1_Deg0(
                knotsx_mult=knotsx_mult,
                knotsy_mult=knotsy_mult,
                cropbs=cropbs,
                cropbs_flat=cropbs_flat,
                nx=nx,
                ny=ny,
                nbs=nbs,
            )

            # surface elements
            dZ = np.repeat(knotsy_mult[1:] - knotsy_mult[:-1], nx)
            if geometry == 'linear':
                dR = np.tile(
                    knotsx_mult[1:] - knotsx_mult[:-1],
                    ny,
                )
            else:
                dR = np.tile(
                    0.5*(knotsx_mult[1:]**2 - knotsx_mult[:-1]**2),
                    ny,
                )

            if cropbs_flat is not False:
                dR = dR[cropbs_flat]
                dZ = dZ[cropbs_flat]

            opmat = (
                scpsp.csc_matrix(gradR.T.dot(gradR*(dR*dZ)[:, None])),
                scpsp.csc_matrix(gradZ.T.dot(gradZ*(dR*dZ)[:, None])),
            )

        else:

            # pre-compute integrals for exact operator deg >= 1
            if deg == 1:
                idR = _D1N2_Deg1(knotsx_mult, geometry=geometry)
                idZ = _D1N2_Deg1(knotsy_mult, geometry='linear')
                iR = _D0N2_Deg1(knotsx_mult, geometry=geometry)
                iZ = _D0N2_Deg1(knotsy_mult, geometry='linear')
            elif deg == 2:
                idR = _D1N2_Deg2(knotsx_mult, geometry=geometry)
                idZ = _D1N2_Deg2(knotsy_mult, geometry='linear')
                iR = _D0N2_Deg2(knotsx_mult, geometry=geometry)
                iZ = _D0N2_Deg2(knotsy_mult, geometry='linear')
            elif deg == 3:
                msg = "Integral D1N2 not implemented for deg=3!"
                raise NotImplementedError(msg)

            # set non-diagonal elements
            i0 = 0
            for ir in range(nx):
                for iz in range(ny):

                    iflat = ir + iz*nx
                    if cropbs_flat is not False and not cropbs_flat[iflat]:
                        continue

                    # general case
                    overlapi = overlap[:, iflat][overlap[:, iflat] > iflat]

                    # diagonal element
                    datadR[i0] = idR[0, ir] * iZ[0, iz]
                    datadZ[i0] = iR[0, ir] * idZ[0, iz]
                    row[i0] = indbs[iflat]
                    column[i0] = indbs[iflat]
                    i0 += 1

                    # non-diagonal elements (symmetric)
                    for jflat in overlapi:

                        if cropbs_flat is not False and not cropbs_flat[jflat]:
                            continue

                        jr = jflat % nx
                        jz = jflat // nx

                        # store (i, j) and (j, i) (symmetric matrix)
                        if jr >= ir:
                            iidR = idR[jr - ir, ir]
                            iiR = iR[jr - ir, ir]
                        else:
                            iidR = idR[abs(jr - ir), jr]
                            iiR = iR[abs(jr - ir), jr]
                        if jz >= iz:
                            iidZ = idZ[jz - iz, iz]
                            iiZ = iZ[jz - iz, iz]
                        else:
                            iidZ = idZ[abs(jz - iz), jz]
                            iiZ = iZ[abs(jz - iz), jz]
                        datadR[i0:i0+2] = iidR * iiZ
                        datadZ[i0:i0+2] = iiR * iidZ
                        row[i0:i0+2] = (indbs[iflat], indbs[jflat])
                        column[i0:i0+2] = (indbs[jflat], indbs[iflat])
                        i0 += 2

            assert i0 == nbtot
            opmat = (
                scpsp.csc_matrix((datadR, (row, column)), shape=shape),
                scpsp.csc_matrix((datadZ, (row, column)), shape=shape),
            )

    # ------------
    # D2N2

    elif operator == 'D2N2':

        # pre-compute integrals
        if deg == 2:
            id2R = _D2N2_Deg2(knotsx_mult, geometry=geometry)
            id2Z = _D2N2_Deg2(knotsy_mult, geometry='linear')
            idR = _D1N2_Deg2(knotsx_mult, geometry=geometry)
            idZ = _D1N2_Deg2(knotsy_mult, geometry='linear')
            iR = _D0N2_Deg2(knotsx_mult, geometry=geometry)
            iZ = _D0N2_Deg2(knotsy_mult, geometry='linear')
        elif deg == 3:
            msg = "Integral D2N2 not implemented for deg=3!"
            raise NotImplementedError(msg)

        # set non-diagonal elements
        i0 = 0
        for ir in range(nx):
            for iz in range(ny):

                iflat = ir + iz*nx
                if cropbs_flat is not False and not cropbs_flat[iflat]:
                    continue

                # general case
                overlapi = overlap[:, iflat][overlap[:, iflat] > iflat]

                # diagonal element
                datad2R[i0] = id2R[0, ir] * iZ[0, iz]
                datad2Z[i0] = iR[0, ir] * id2Z[0, iz]
                datadRZ[i0] = idR[0, ir] * idZ[0, iz]
                row[i0] = indbs[iflat]
                column[i0] = indbs[iflat]
                i0 += 1

                # non-diagonal elements (symmetric)
                for jflat in overlapi:

                    if cropbs_flat is not False and not cropbs_flat[jflat]:
                        continue

                    jr = jflat % nx
                    jz = jflat // nx

                    # store (i, j) and (j, i) (symmetric matrix)
                    if jr >= ir:
                        iid2R = id2R[jr - ir, ir]
                        iidR = idR[jr - ir, ir]
                        iiR = iR[jr - ir, ir]
                    else:
                        iid2R = id2R[abs(jr - ir), jr]
                        iidR = idR[abs(jr - ir), jr]
                        iiR = iR[abs(jr - ir), jr]
                    if jz >= iz:
                        iid2Z = id2Z[jz - iz, iz]
                        iidZ = idZ[jz - iz, iz]
                        iiZ = iZ[jz - iz, iz]
                    else:
                        iid2Z = id2Z[abs(jz - iz), jz]
                        iidZ = idZ[abs(jz - iz), jz]
                        iiZ = iZ[abs(jz - iz), jz]
                    datad2R[i0:i0+2] = iid2R * iiZ
                    datad2Z[i0:i0+2] = iiR * iid2Z
                    datadRZ[i0:i0+2] = iidR * iidZ
                    row[i0:i0+2] = (indbs[iflat], indbs[jflat])
                    column[i0:i0+2] = (indbs[jflat], indbs[iflat])
                    i0 += 2

        assert i0 == nbtot
        opmat = (
            scpsp.csc_matrix((datad2R, (row, column)), shape=shape),
            scpsp.csc_matrix((datad2Z, (row, column)), shape=shape),
            scpsp.csc_matrix((datadRZ, (row, column)), shape=shape),
        )

    # ------------
    # D3N2

    elif operator == 'D3N2' and deg == 3:

        raise NotImplementedError("Integral D3N2 not implemented for deg=3!")

    return opmat, operator, geometry, dim


# #############################################################################
# #############################################################################
#               Operator sub-routines: D0N2
# #############################################################################


def _D0N2_Deg1_full_linear(k0, k2):
    """ from 1d knots, return int_0^2 x b**2(x) dx """
    return (k2 - k0) / 3.


def _D0N2_Deg1_full_toroidal(k0, k1, k2):
    """ from 1d knots, return int_0^2 x b**2(x) dx """
    intt = np.zeros((k0.size,))
    intt[1:] += (
        (3. * k1**3 - 5.*k0*k1**2 + k1*k0**2 + k0**3)[1:]
        / (12. * (k1 - k0))[1:]
    )
    intt[:-1] = (
        + (3.*k1**3 - 5.*k2*k1**2 + k1*k2**2 + k2**3)[:-1]
        / (12. * (k2 - k1))[:-1]
    )
    return intt


def _D0N2_Deg1_2_linear(k1, k2):
    """ from 1d knots, return int_0^2 x b**2(x) dx """
    return (k2 - k1) / 6.


def _D0N2_Deg1_2_toroidal(k1, k2):
    """ from 1d knots, return int_0^2 x b**2(x) dx """
    return (k2**2 - k1**2) / 12.


def _D0N2_Deg1(knots, geometry=None):

    if geometry == 'linear':
        integ = np.array([
            _D0N2_Deg1_full_linear(
                knots[:-2],
                knots[2:],
            ),
            _D0N2_Deg1_2_linear(
                knots[1:-1],
                knots[2:]
            ),
        ])
    else:
        integ = np.array([
            _D0N2_Deg1_full_toroidal(
                knots[:-2],
                knots[1:-1],
                knots[2:],
            ),
            _D0N2_Deg1_2_toroidal(
                knots[1:-1],
                knots[2:]
            ),
        ])
    return integ


def _D0N2_Deg2(knots, geometry=None):

    if geometry == 'linear':
        ffull = _D0N2_Deg2_full_linear
        f3 = _D0N2_Deg2_3_linear
        f2 = _D0N2_Deg2_2_linear
    else:
        ffull = _D0N2_Deg2_full_toroidal
        f3 = _D0N2_Deg2_3_toroidal
        f2 = _D0N2_Deg2_2_toroidal

    integ = np.array([
        ffull(
            knots[:-3],
            knots[1:-2],
            knots[2:-1],
            knots[3:],
        ),
        f3(
            knots[:-3],
            knots[1:-2],
            knots[2:-1],
            knots[3:],
            np.r_[knots[4:], np.nan],
        ),
        f2(
            knots[1:-2],
            knots[2:-1],
            knots[3:],
            np.r_[knots[4:], np.nan],
        ),
    ])
    return integ


def _D0N2_Deg2_full_linear(k0, k1, k2, k3):
    """ from 1d knots, return int_0^3 b**2(x) dx """
    intt = np.zeros((k0.size,))
    intt[1:] += (
        (k1 - k0)[1:]**3 / (5.*(k2 - k0)[1:]**2)
        + (k2 - k1)[1:]
        * (
            10.*k0**2 + 6.*k1**2 + 3.*k1*k2 + k2**2 - 5.*k0*(3.*k1 + k2)
        )[1:] / (30.*(k2 - k0)[1:]**2)
    )
    intt[1:-1] += (
        (k2 - k1)[1:-1]
        * (
            -3.*k1**2 - 4.*k1*k2 - 3.*k2**2 + 5.*k0*(k1 + k2 - 2.*k3)
            + 5.*k3*(k1 + k2)
        )[1:-1] / (60.*(k2 - k0)*(k3 - k1))[1:-1]
    )
    intt[:-1] += (
        (k2 - k1)[:-1]
        * (
            10.*k3**2 + 6.*k2**2 + 3.*k1*k2 + k1**2 - 5.*k3*(3.*k2 + k1)
        )[:-1] / (30.*(k3 - k1)[:-1]**2)
        + (k3 - k2)[:-1]**3 / (5.*(k3 - k1)[:-1]**2)
    )
    return intt


def _D0N2_Deg2_full_toroidal(k0, k1, k2, k3):
    """ from 1d knots, return int_0^3 b**2(x) dx """
    intt = np.zeros((k0.size,))
    intt[1:] += (
        (5.*k1 + k0)[1:]*(k1 - k0)[1:]**3 / (30.*(k2 - k0)[1:]**2)
        + (k2 - k1)[1:]
        * (
            10*k1**3 + 6.*k1**2*k2 + 3.*k1*k2**2
            + k2**3 + 5.*k0**2*(3.*k1 + k2)
            - 4.*k0*(6.*k1**2 + 3.*k1*k2 + k2**2)
        )[1:] / (60.*(k2 - k0)**2)[1:]
    )
    intt[:-1] += (
        (5.*k2 + k3)[:-1]*(k3 - k2)[:-1]**3 / (30.*(k3 - k1)[:-1]**2)
        + (k2 - k1)[1:]
        * (
            10*k2**3 + 6.*k2**2*k1 + 3.*k2*k1**2
            + k1**3 + 5.*k3**2*(3.*k2 + k1)
            - 4.*k3*(6.*k2**2 + 3.*k2*k1 + k1**2)
        )[:-1] / (60.*(k3 - k1)**2)[:-1]
    )
    intt[1:-1] += (
        (k2 - k1)[1:-1]
        * (
            - 2.*k1**3 - 2.*k2**3
            - 3.*k1*k2*(k1 + k2)
            - 5.*k0*k3*(k1 + k2)
            + (k0 + k3)*(3.*k2**2 + 4.*k1*k2 + 3.*k1**2)
        )[1:-1] / (30.*(k3 - k1)*(k2 - k0))[1:-1]
    )

    return intt


def _D0N2_Deg2_3_linear(k0, k1, k2, k3, k4):
    """ from 1d knots, return int_0^3 b**2(x) dx """
    intt = np.zeros((k0.size,))
    intt[1:-1] += (
        (3.*k2 + 2.*k1 - 5.*k0)[1:-1]*(k2 - k1)[1:-1]**2
        / (60.*(k3 - k1)*(k2 - k0))[1:-1]
    )
    intt[:-2] += (
        + (5.*k4 - 2.*k3 - 3.*k2)[:-2]*(k3 - k2)[:-2]**2
        / (60.*(k4 - k2)*(k3 - k1))[:-2]
    )
    intt[:-1] += (
        + (5.*k3 - 4.*k2 - k1)[:-1]*(k2 - k1)[:-1]**2
        / (20.*(k3 - k1)**2)[:-1]
        + (4.*k2 + k3 - 5.*k1)[:-1]*(k3 - k2)[:-1]**2
        / (20.*(k3 - k1)**2)[:-1]
    )
    return intt


def _D0N2_Deg2_3_toroidal(k0, k1, k2, k3, k4):
    """ from 1d knots, return int_0^3 b**2(x) dx """
    intt = np.zeros((k0.size,))
    intt[:-1] = (
        (k2 - k1)[:-1]**2
        * (-10*k2**2 - 4*k1*k2 - k1**2 + 3*k3*(4*k2 + k1))[:-1]
        / (60.*(k3 - k1)**2)[:-1]
        + (k3 - k2)[:-1]**2
        * (k3**2 + 4*k3*k2 + 10*k2**2 - 3*k1*(k3 + 4*k2))[:-1]
        / (60*(k3 - k1)**2)[:-1]
    )
    intt[1:-1] = (
        (k2 - k1)[1:-1]**2
        * (2*k2**2 + 2*k1*k2 + k1**2 - k0*(3.*k2 + 2.*k1))[1:-1]
        / (60.*(k3 - k1)*(k2 - k0))[1:-1]
    )
    intt[:-2] = (
        + (k3 - k2)[:-2]**2
        * (-k3**2 - 2*k3*k2 - 2*k2**2 + k4*(2*k3 + 3*k2))[:-2]
        / (60*(k4 - k2)*(k3 - k1))[:-2]
    )
    return intt


def _D0N2_Deg2_2_linear(k1, k2, k3, k4):
    """ from 1d knots, return int_0^3 b**2(x) dx """
    intt = np.zeros((k1.size,))
    intt[:-2] = (
        (k3 - k2)[:-2]**3
        / (30.*(k4 - k2)*(k3 - k1))[:-2]
    )
    return intt


def _D0N2_Deg2_2_toroidal(k1, k2, k3, k4):
    """ from 1d knots, return int_0^3 b**2(x) dx """
    intt = np.zeros((k1.size,))
    intt[:-2] = (
        (k3 + k2)[:-2]*(k3 - k2)[:-2]**3
        / (60.*(k4 - k2)[:-2]*(k3 - k1)[:-2])
    )
    return intt


# #############################################################################
# #############################################################################
#               Operator sub-routines: D1 - deg = 0 - discrete
# #############################################################################


def _D1_Deg0(
    knotsx_mult=None,
    knotsy_mult=None,
    cropbs=None,
    cropbs_flat=None,
    nx=None,
    ny=None,
    nbs=None,
):
    """ Discrete apprmixation of the gradient for pixels

    Centered when possible
    Non-centered otherwise

    """

    # initialize output
    datadR = np.zeros((nbs, nbs), dtype=float)
    datadZ = np.zeros((nbs, nbs), dtype=float)
    if cropbs is False:
        cropbs = np.ones((nx, ny), dtype=bool)

    # positions of centers
    centsR = 0.5*(knotsx_mult[1:] + knotsx_mult[:-1])
    centsZ = 0.5*(knotsy_mult[1:] + knotsy_mult[:-1])

    # Determine points that have 2 neighbours in R (centered)
    n2R = np.zeros(cropbs.shape, dtype=bool)
    n2R[1:-1, :] = cropbs[1:-1, :] & cropbs[2:, :] & cropbs[:-2, :]
    # points with a neighbours at higher R
    npR = cropbs & (~n2R)
    npR[-1, :] = False
    npR[:-1, :] &= cropbs[1:, :]
    # points with a neighbours at lower R
    nmR = cropbs & (~n2R) & (~npR)
    nmR[0, :] = False
    nmR[1:, :] &= cropbs[:-1, :]

    # Determine points that have 2 neighbours in Z (centered)
    n2Z = np.zeros(cropbs.shape, dtype=bool)
    n2Z[:, 1:-1] = cropbs[:, 1:-1] & cropbs[:, 2:] & cropbs[:, :-2]
    n2Z[:, 1:-1] = n2Z[:, 1:-1] & n2Z[:, 2:] & n2Z[:, :-2]
    # points with a neighbours at higher Z
    npZ = cropbs & (~n2Z)
    npZ[:, -1] = False
    npZ[:, :-1] &= cropbs[:, 1:]
    # points with a neighbours at lower Z
    nmZ = cropbs & (~n2Z) & (~npZ)
    nmZ[:, 0] = False
    nmZ[:, 1:] &= cropbs[:, :-1]

    # iterate on each type of point in R
    for ir, iz in zip(*n2R.nonzero()):
        iflat = ir + iz*nx
        dRi = 1./(centsR[ir + 1] - centsR[ir - 1])
        datadR[iflat, iflat - 1] = -dRi
        datadR[iflat, iflat + 1] = dRi
    for ir, iz in zip(*npR.nonzero()):
        iflat = ir + iz*nx
        dRi = 1./(centsR[ir + 1] - centsR[ir])
        datadR[iflat, iflat] = -dRi
        datadR[iflat, iflat + 1] = dRi
    for ir, iz in zip(*nmR.nonzero()):
        iflat = ir + iz*nx
        dRi = 1./(centsR[ir] - centsR[ir-1])
        datadR[iflat, iflat - 1] = -dRi
        datadR[iflat, iflat] = dRi

    # iterate on each type of point in Z
    for ir, iz in zip(*n2Z.nonzero()):
        iflat = ir + iz*nx
        dZi = 1./(centsZ[iz + 1] - centsZ[iz - 1])
        datadZ[iflat, iflat - nx] = -dZi
        datadZ[iflat, iflat + nx] = dZi
    for ir, iz in zip(*npZ.nonzero()):
        iflat = ir + iz*nx
        dZi = 1./(centsZ[iz + 1] - centsZ[iz])
        datadZ[iflat, iflat] = -dZi
        datadZ[iflat, iflat + nx] = dZi
    for ir, iz in zip(*nmZ.nonzero()):
        iflat = ir + iz*nx
        dZi = 1./(centsZ[iz] - centsZ[iz - 1])
        datadZ[iflat, iflat - nx] = -dZi
        datadZ[iflat, iflat] = dZi

    # crop and return
    if cropbs_flat is False:
        return datadR, datadZ
    else:
        return (
            datadR[cropbs_flat, :][:, cropbs_flat],
            datadZ[cropbs_flat, :][:, cropbs_flat],
        )


# #############################################################################
# #############################################################################
#               Operator sub-routines: D1N2 - exact
# #############################################################################


def _D1N2_Deg1(knots, geometry=None):

    if geometry == 'linear':
        ffull = _D1N2_Deg1_full_linear
        f2 = _D1N2_Deg1_2_linear
    else:
        ffull = _D1N2_Deg1_full_toroidal
        f2 = _D1N2_Deg1_2_toroidal

    integ = np.array([
        ffull(
            knots[:-2],
            knots[1:-1],
            knots[2:],
        ),
        f2(
            knots[1:-1],
            knots[2:]
        ),
    ])
    return integ


def _D1N2_Deg1_full_linear(k0, k1, k2):
    intt = np.zeros((k0.size,))
    intt[1:] += 1. / (k1 - k0)[1:]
    intt[:-1] += 1. / (k2 - k1)[:-1]
    return intt


def _D1N2_Deg1_full_toroidal(k0, k1, k2):
    intt = np.zeros((k0.size,))
    intt[1:] += (k1 + k0)[1:] / (2.*(k1 - k0))[1:]
    intt[:-1] += (k2 + k1)[:-1] / (2.*(k2 - k1))[:-1]
    return intt


def _D1N2_Deg1_2_linear(k1, k2):
    intt = np.zeros((k1.size,))
    intt[:-1] = -1. / (k2 - k1)[:-1]
    return intt


def _D1N2_Deg1_2_toroidal(k1, k2):
    intt = np.zeros((k1.size,))
    intt[:-1] = - (k2 + k1)[:-1] / (2.*(k2 - k1))[:-1]
    return intt


def _D1N2_Deg2(knots, geometry=None):

    if geometry == 'linear':
        ffull = _D1N2_Deg2_full_linear
        f3 = _D1N2_Deg2_3_linear
        f2 = _D1N2_Deg2_2_linear
    else:
        ffull = _D1N2_Deg2_full_toroidal
        f3 = _D1N2_Deg2_3_toroidal
        f2 = _D1N2_Deg2_2_toroidal

    integ = np.array([
        ffull(
            knots[:-3],
            knots[1:-2],
            knots[2:-1],
            knots[3:],
        ),
        f3(
            knots[:-3],
            knots[1:-2],
            knots[2:-1],
            knots[3:],
            np.r_[knots[4:], np.nan],
        ),
        f2(
            knots[1:-2],
            knots[2:-1],
            knots[3:],
            np.r_[knots[4:], np.nan],
        ),
    ])
    return integ


def _D1N2_Deg2_full_linear(k0, k1, k2, k3):
    intt = np.zeros((k0.size,))
    intt[1:] += 4.*(k1 - k0)[1:] / (3.*(k2 - k0)[1:]**2)
    intt[:-1] += 4.*(k3 - k2)[:-1] / (3.*(k3 - k1)[:-1]**2)
    intt[1:-1] += (
        4.*(k2 - k1)[1:-1]
        * (
            k2**2 + k2*k1 + k1**2 + k3**2 + k0*k3 + k0**2
            - k3*(k2 + 2.*k1) - k0*(2.*k2 + k1)
        )[1:-1]
        / (3.*(k3 - k1)[1:-1]**2*(k2 - k0)[1:-1]**2)
    )
    return intt


def _D1N2_Deg2_full_toroidal(k0, k1, k2, k3):
    intt = np.zeros((k0.size,))
    intt[1:] += (3.*k1 + k0)[1:]*(k1 - k0)[1:] / (3.*(k2 - k0)[1:]**2)
    intt[:-1] += (3.*k2 + k3)[:-1]*(k3 - k2)[:-1] / (3.*(k3 - k1)[:-1]**2)
    intt[1:-1] += (
        (k2 - k1)[1:-1]
        * (
            3.*(k2 + k1)*(k2**2 + k1**2)
            + k3**2*(k2 + 3.*k1)
            + k0**2*(3.*k2 + k1)
            - 2.*k3*(k2**2 + 2.*k2*k1 + 3.*k1**2)
            - 2.*k0*(3.*k2**2 + 2.*k2*k1 + k1**2)
            + 2.*k3*k0*(k2 + k1)
        )[1:-1]
        / (3.*(k3 - k1)[1:-1]**2*(k2 - k0)[1:-1]**2)
    )
    return intt


def _D1N2_Deg2_3_linear(k0, k1, k2, k3, k4):
    intt = np.zeros((k0.size,))
    intt[1:-1] += (
        2.*(k2 - k1)[1:-1]
        * (k3 - 2.*k2 - k1 + 2.*k0)[1:-1]
        / (3.*(k3 - k1)**2*(k2 - k0))[1:-1]
    )
    intt[:-2] += (
        2.*(k3 - k2)[:-2]
        * (-2.*k4 + k3 + 2.*k2 - k1)[:-2]
        / (3.*(k4 - k2)*(k3 - k1)**2)[:-2]
    )
    return intt


def _D1N2_Deg2_3_toroidal(k0, k1, k2, k3, k4):
    intt = np.zeros((k0.size,))
    intt[1:-1] += (
        (k2 - k1)[1:-1]
        * (
            - (3.*k2**2 + 2.*k2*k1 + k1**2)
            + k3*(k2 + k1)
            + k0*(3.*k2 + k1)
        )[1:-1]
        / (3.*(k3 - k1)**2*(k2 - k0))[1:-1]
    )
    intt[:-2] += (
        (k3 - k2)[:-2]
        * (
            k3**2 + 2.*k2*k3 + 3.*k2**2
            - k4*(k3 + 3.*k2)
            - k1*(k3 + k2)
        )[:-2]
        / (3.*(k4 - k2)*(k3 - k1)**2)[:-2]
    )
    return intt


def _D1N2_Deg2_2_linear(k1, k2, k3, k4):
    intt = np.zeros((k1.size,))
    intt[:-2] += -(
        2.*(k3 - k2)[:-2]
        / (3.*(k4 - k2)*(k3 - k1))[:-2]
    )
    return intt


def _D1N2_Deg2_2_toroidal(k1, k2, k3, k4):
    intt = np.zeros((k1.size,))
    intt[:-2] += -(
        (k3 + k2)[:-2]*(k3 - k2)[:-2]
        / (3.*(k4 - k2)*(k3 - k1))[:-2]
    )
    return intt


# #############################################################################
# #############################################################################
#               Operator sub-routines: D2N2
# #############################################################################


def _D2N2_Deg2(knots, geometry=None):

    if geometry == 'linear':
        ffull = _D2N2_Deg2_full_linear
        f3 = _D2N2_Deg2_3_linear
        f2 = _D2N2_Deg2_2_linear
    else:
        ffull = _D2N2_Deg2_full_toroidal
        f3 = _D2N2_Deg2_3_toroidal
        f2 = _D2N2_Deg2_2_toroidal

    integ = np.array([
        ffull(
            knots[:-3],
            knots[1:-2],
            knots[2:-1],
            knots[3:],
        ),
        f3(
            knots[:-3],
            knots[1:-2],
            knots[2:-1],
            knots[3:],
            np.r_[knots[4:], np.nan],
        ),
        f2(
            knots[1:-2],
            knots[2:-1],
            knots[3:],
            np.r_[knots[4:], np.nan],
        ),
    ])
    return integ


def _D2N2_Deg2_full_linear(k0, k1, k2, k3):
    intt = np.zeros((k0.size,))
    intt[2:] += 4. / ((k2 - k0)**2*(k1 - k0))[2:]
    intt[:-2] += 4. / ((k3 - k2)*(k3 - k1)**2)[:-2]
    intt[1:-1] += (
        4.*(k3 + k2 - k1 - k0)[1:-1]**2
        / ((k3 - k1)**2*(k2 - k1)*(k2 - k0)**2)[1:-1]
    )
    return intt


def _D2N2_Deg2_full_toroidal(k0, k1, k2, k3):
    intt = np.zeros((k0.size,))
    intt[2:] += 2.*(k0 + k0)[2:] / ((k2 - k0)**2*(k1 - k0))[2:]
    intt[:-2] += 2.*(k3 + k2)[2:] / ((k3 - k2)*(k3 - k1)**2)[:-2]
    intt[1:-1] += (
        (2.*(k3 + k2 - k1 - k0)**2*(k2 + k1))[1:-1]
        / ((k3 - k1)**2*(k2 - k1)*(k2 - k0)**2)[1:-1]
    )
    return intt


def _D2N2_Deg2_3_linear(k0, k1, k2, k3, k4):
    intt = np.zeros((k0.size,))
    intt[1:-1] += (
        - 4.*(k3 + k2 - k1 - k0)[1:-1]
        / ((k3 - k1)**2*(k2 - k1)*(k2 - k0))[1:-1]
    )
    intt[:-2] += (
        - 4.*(k4 + k3 - k2 - k1)[:-2]
        / ((k4 - k2)*(k3 - k2)*(k3 - k1)**2)[:-2]
    )
    return intt


def _D2N2_Deg2_3_toroidal(k0, k1, k2, k3, k4):
    intt = np.zeros((k0.size,))
    intt[1:-1] += (
        - 2.*((k3 + k2 - k1 - k0)*(k2 + k1))[1:-1]
        / ((k3 - k1)**2*(k2 - k1)*(k2 - k0))[1:-1]
    )
    intt[:-2] += (
        - 2.*((k4 + k3 - k2 - k1)*(k3 + k2))[:-2]
        / ((k4 - k2)*(k3 - k2)*(k3 - k1)**2)[:-2]
    )
    return intt


def _D2N2_Deg2_2_linear(k1, k2, k3, k4):
    intt = np.zeros((k1.size,))
    intt[:-2] += 4. / ((k4 - k2)*(k3 - k2)*(k3 - k1))[:-2]
    return intt


def _D2N2_Deg2_2_toroidal(k1, k2, k3, k4):
    intt = np.zeros((k1.size,))
    intt[:-2] += 2.*(k3 + k2)[:-2] / ((k4 - k2)*(k3 - k2)*(k3 - k1))[:-2]
    return intt



# -*- coding: utf-8 -*-


import copy
import itertools as itt


import numpy as np
import scipy.interpolate as scpinterp
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors


import datastock as ds


from . import _utils_surface3d


# ##################################################################
# ##################################################################
#                   optics outline
# ##################################################################


def get_optics_outline(
    coll=None,
    key=None,
    add_points=None,
    mode=None,
    closed=None,
    ravel=None,
    total=None,
):

    # ------------
    # check inputs

    # key, cls
    key, cls = coll.get_optics_cls(optics=key)
    key, cls = key[0], cls[0]
    dgeom = coll.dobj[cls][key]['dgeom']

    # total
    total = ds._generic_check._check_var(
        total, 'total',
        types=bool,
        default=(cls == 'camera' and dgeom['type'] == '2d'),
    )
    if cls == 'camera' and dgeom['type'] != '2d':
        total = False

    # --------
    # compute

    if dgeom['type'] == '3d':
        return None, None

    if cls == 'camera' and total:
        # get centers
        cx0, cx1 = dgeom['cents']
        cx0 = coll.ddata[cx0]['data']
        cx1 = coll.ddata[cx1]['data']

        # derive half-spacing
        dx0 = np.mean(np.diff(cx0)) / 2.
        dx1 = np.mean(np.diff(cx1)) / 2.

        # derive global outline (not pixel outline)
        p0 = np.r_[
            cx0[0] - dx0, cx0[-1] + dx0,
            cx0[-1] + dx0, cx0[0] - dx0,
        ]
        p1 = np.r_[
            cx1[0] - dx1, cx1[0] - dx1,
            cx1[-1] + dx1, cx1[-1] + dx1,
        ]

    else:
        out = dgeom['outline']
        p0 = coll.ddata[out[0]]['data']
        p1 = coll.ddata[out[1]]['data']

    # -----------
    # add_points

    return _interp_poly(
        lp=[p0, p1],
        add_points=add_points,
        mode=mode,
        isclosed=False,
        closed=closed,
        ravel=ravel,
    )


# ##################################################################
# ##################################################################
#                   optics poly
# ##################################################################


def get_optics_poly(
    coll=None,
    key=None,
    add_points=None,
    mode=None,
    closed=None,
    ravel=None,
    total=None,
    return_outline=None,
):

    # ------------
    # check inputs

    key, cls = coll.get_optics_cls(optics=key)
    key, cls = key[0], cls[0]

    return_outline = ds._generic_check._check_var(
        return_outline, 'return_outline',
        types=bool,
        default=False,
    )

    ravel = ds._generic_check._check_var(
        ravel, 'ravel',
        default=False,
        types=bool,
    )

    # --------
    # compute

    dgeom = coll.dobj[cls][key]['dgeom']
    if cls in ['aperture', 'filter', 'crystal', 'grating']:

        if dgeom['type'] != '3d':
            p0, p1 = coll.get_optics_outline(
                key=key,
                add_points=add_points,
                mode=mode,
                closed=closed,
                ravel=ravel,
                total=total,
            )

            px, py, pz = _utils_surface3d._get_curved_poly(
                gtype=dgeom['type'],
                outline_x0=p0,
                outline_x1=p1,
                curve_r=dgeom['curve_r'],
                cent=dgeom['cent'],
                nin=dgeom['nin'],
                e0=dgeom['e0'],
                e1=dgeom['e1'],
            )

        else:
            px, py, pz = dgeom['poly']
            px = coll.ddata[px]['data']
            py = coll.ddata[py]['data']
            pz = coll.ddata[pz]['data']

    elif cls == 'camera':

        p0, p1 = coll.get_optics_outline(
            key=key,
            add_points=add_points,
            mode=mode,
            closed=closed,
            ravel=ravel,
            total=total,
        )

        # vectors
        dv = coll.get_camera_unit_vectors(key)
        lv = ['e0_x', 'e0_y', 'e0_z', 'e1_x', 'e1_y', 'e1_z']
        e0x, e0y, e0z, e1x, e1y, e1z = [dv[k0] for k0 in lv]
        if not np.isscalar(e0x):
            e0x = e0x[:, None]
            e0y = e0y[:, None]
            e0z = e0z[:, None]
            e1x = e1x[:, None]
            e1y = e1y[:, None]
            e1z = e1z[:, None]

        if dgeom['type'] == '2d' and total:
            cx, cy, cz = dgeom['cent']
            p02, p12 = p0, p1
        else:
            cx, cy, cz = coll.get_camera_cents_xyz(key)
            shape = [1 for ii in range(cx.ndim)] + [p0.size]
            cx, cy, cz = cx[..., None], cy[..., None], cz[..., None]
            p02 = p0.reshape(shape)
            p12 = p1.reshape(shape)

        # make 3d
        px = cx + p02 * e0x + p12 * e1x
        py = cy + p02 * e0y + p12 * e1y
        pz = cz + p02 * e0z + p12 * e1z

    # ----------
    # ravel

    if ravel is True and px.ndim > 1:
        nan = np.full(tuple(np.r_[px.shape[:-1], 1]), np.nan)
        px = np.concatenate((px, nan), axis=-1).ravel()
        py = np.concatenate((py, nan), axis=-1).ravel()
        pz = np.concatenate((pz, nan), axis=-1).ravel()

    # return
    if return_outline is True:
        return p0, p1, px, py, pz
    else:
        return px, py, pz


# ##################################################################
# ##################################################################
#                   Poly interpolation utilities
# ##################################################################


def _interp_poly_check(
    add_points=None,
    mode=None,
    closed=None,
    ravel=None,
):

    # -------
    # mode

    mode = ds._generic_check._check_var(
        mode, 'mode',
        default=None,
        allowed=[None, 'min'],
    )

    # ----------
    # add_points

    defadd = 1 if mode == 'min' else 0
    add_points = ds._generic_check._check_var(
        add_points, 'add_points',
        types=int,
        default=defadd,
        sign='>= 0',
    )

    # -------
    # closed

    closed = ds._generic_check._check_var(
        closed, 'closed',
        default=False,
        types=bool,
    )

    # -------
    # ravel

    ravel = ds._generic_check._check_var(
        ravel, 'ravel',
        default=False,
        types=bool,
    )
    return add_points, mode, closed, ravel


def _interp_poly(
    lp=None,
    add_points=None,
    mode=None,
    isclosed=None,
    closed=None,
    ravel=None,
    min_threshold=1.e-6,
):

    # ------------
    # check inputs

    add_points, mode, closed, ravel = _interp_poly_check(
        add_points=add_points,
        mode=mode,
        closed=closed,
        ravel=ravel,
    )

    # ------------
    # trivial case

    if add_points == 0:
        return lp

    # ------------
    # compute

    # close for interpolation
    if isclosed is not True:
        for ii, pp in enumerate(lp):

            if pp is None:
                continue

            if pp.ndim == 2:
                lp[ii] = np.concatenate((pp, pp[:, 0:1]), axis=1)
            else:
                lp[ii] = np.append(pp, pp[0])

    # -----------
    # mode

    if mode == 'min':
        if len(lp) == 3:
            dist = np.sqrt(
                np.diff(lp[0], axis=-1)**2
                + np.diff(lp[1], axis=-1)**2
                + np.diff(lp[2], axis=-1)**2
            )
        elif len(lp) == 2:
            dist = np.sqrt(
                np.diff(lp[0], axis=-1)**2
                + np.diff(lp[1], axis=-1)**2
            )

        if dist.ndim == 2:
            import pdb; pdb.set_trace()     # DB

        min_threshold = min(min_threshold, np.max(dist)/3.)
        mindist = np.min(dist[dist > min_threshold])
        add_points = add_points * np.ceil(dist / mindist).astype(int) - 1

    # -----------
    # add_points

    shape = [pp for pp in lp if pp is not None][0].shape
    nb = shape[-1]
    if np.isscalar(add_points):
        add_points = np.full((nb-1,), add_points, dtype=int)

    # -----------
    # indices

    ind0 = np.arange(0, nb)
    ind = np.concatenate(tuple([
        np.linspace(
            ind0[ii],
            ind0[ii+1],
            2 + add_points[ii],
            endpoint=True,
        )[:-1]
        for ii in range(nb-1)
    ] + [[ind0[-1]]]))

    # -----------
    # interpolate

    for ii, pp in enumerate(lp):

        if pp is None:
            continue

        lp[ii] = scpinterp.interp1d(
            ind0, pp, kind='linear', axis=-1,
        )(ind)

    # ------------
    # closed

    if closed is False:

        for ii, pp in enumerate(lp):
            if pp is None:
                continue

            if pp.ndim == 2:
                lp[ii] = pp[:, :-1]
            else:
                lp[ii] = pp[:-1]

    # ------------
    # ravel

    if ravel and len(shape) == 2:
        nan = np.full((pp.shape[0], 1), np.nan)
        for ii, pp in enumerate(lp[2:]):
            lp[ii+2] = np.concatenate((pp, nan), axis=1).ravel()
    return lp


# ##################################################################
# ##################################################################
#                       dplot
# ##################################################################


def _dplot_check(
    coll=None,
    key=None,
    key_cam=None,
    optics=None,
    elements=None,
    vect_length=None,
    axis_length=None,
):
    # -----
    # key

    key, key_cam = coll.get_diagnostic_cam(key, key_cam)

    # ------
    # optics

    if isinstance(optics, str):
        optics = [optics]

    lok = list(itt.chain.from_iterable([
        [k0] + v0['optics']
        for k0, v0 in coll.dobj['diagnostic'][key]['doptics'].items()
    ]))
    optics = ds._generic_check._check_var_iter(
        optics, 'optics',
        default=lok,
        allowed=lok,
    )

    # -------
    # elements

    lok = ['o', 'c', 'v', 'r']
    elements = ds._generic_check._check_var_iter(
        elements, 'elements',
        types=str,
        types_iter=str,
        default=''.join(lok),
        allowed=lok,
    )

    # -----------
    # vect_length

    vect_length = ds._generic_check._check_var(
        vect_length, 'vect_length',
        default=0.2,
        types=(float, int),
        sign='>= 0.'
    )

    # -----------
    # axis_length

    axis_length = ds._generic_check._check_var(
        axis_length, 'axis_length',
        default=1.,
        types=(float, int),
        sign='>= 0.'
    )

    return key, key_cam, optics, elements, vect_length, axis_length


def _dplot(
    coll=None,
    key=None,
    key_cam=None,
    optics=None,
    elements=None,
    vect_length=None,
    axis_length=None,
):

    # ------------
    # check inputs

    key, key_cam, optics, elements, vect_length, axis_length = _dplot_check(
        coll=coll,
        key=key,
        key_cam=key_cam,
        optics=optics,
        elements=elements,
        vect_length=vect_length,
        axis_length=axis_length,
    )

    # ------------
    # build dict

    dlw = {
        'camera': 2,
        'aperture': 1.,
        'filter': 1.,
        'crystal': 1.,
        'grating': 1.,
    }
    dplot = {k0: {} for k0 in optics}
    for k0 in optics:

        if k0 in coll.dobj.get('camera', []):
            cls = 'camera'
        elif k0 in coll.dobj.get('aperture', []):
            cls = 'aperture'
        elif k0 in coll.dobj.get('filter', []):
            cls = 'filter'
        elif k0 in coll.dobj.get('crystal', []):
            cls = 'crystal'
        elif k0 in coll.dobj.get('grating', []):
            cls = 'grating'
        else:
            msg = f"Unknown optics '{k0}'"
            raise Exception(msg)

        v0 = coll.dobj[cls][k0]['dgeom']
        color = coll.dobj[cls][k0]['dmisc']['color']

        # -----------
        # prepare data

        # cent
        if 'c' in elements or 'v' in elements or 'r' in elements:
            if v0.get('cent') is not None:
                cx, cy, cz = v0['cent'][:, None]
            elif 'cents' in v0.keys():
                cx, cy, cz = v0['cents']
                cx = coll.ddata[cx]['data']
                cy = coll.ddata[cy]['data']
                cz = coll.ddata[cz]['data']
            cr = np.hypot(cx, cy)

        # vectors
        if 'v' in elements or 'r' in elements:
            ninx, niny, ninz = v0['nin']
            e0x, e0y, e0z = v0['e0']
            e1x, e1y, e1z = v0['e1']
            if isinstance(ninx, str):
                vinx = coll.ddata[ninx]['data'] * vect_length
                viny = coll.ddata[niny]['data'] * vect_length
                vinz = coll.ddata[ninz]['data'] * vect_length
                v0x = coll.ddata[e0x]['data'] * vect_length
                v0y = coll.ddata[e0y]['data'] * vect_length
                v0z = coll.ddata[e0z]['data'] * vect_length
                v1x = coll.ddata[e1x]['data'] * vect_length
                v1y = coll.ddata[e1y]['data'] * vect_length
                v1z = coll.ddata[e1z]['data'] * vect_length
            else:
                vinx, viny, vinz = np.r_[ninx, niny, ninz] * vect_length
                v0x, v0y, v0z = np.r_[e0x, e0y, e0z] * vect_length
                v1x, v1y, v1z = np.r_[e1x, e1y, e1z] * vect_length

        # radius
        if 'r' in elements and v0['type'] not in ['planar', '1d', '2d', '3d']:
            if v0['type'] == 'cylindrical':
                icurv = (np.isfinite(v0['curve_r'])).nonzero()[0][0]
                rc = v0['curve_r'][icurv]
                eax = [(e0x, e0y, e0z), (e1x, e1y, e1z)][1 - icurv]
            elif v0['type'] == 'spherical':
                rc = v0['curve_r'][0]
            elif v0['type'] == 'toroidal':
                imax = np.argmax(v0['curve_r'])
                imin = 1 - imax
                rmax = v0['curve_r'][imax]
                rmin = v0['curve_r'][imin]
                emax = [(e0x, e0y, e0z), (e1x, e1y, e1z)][imax]
            # extenthalf = v0['extenthalf']

        # -----------------
        # get plotting data

        # outline
        if 'o' in elements:

            p0, p1, px, py, pz = coll.get_optics_poly(
                key=k0,
                add_points=3,
                closed=True,
                ravel=True,
                return_outline=True,
                total=True,
            )

            dplot[k0]['o'] = {
                'x0': p0,
                'x1': p1,
                'x': px,
                'y': py,
                'z': pz,
                'r': np.hypot(px, py),
                'props': {
                    'label': f'{k0}-o',
                    'lw': dlw[cls],
                    'c': color,
                },
            }

        # center
        if 'c' in elements:

            dplot[k0]['c'] = {
                'x': cx,
                'y': cy,
                'z': cz,
                'r': cr,
                'props': {
                    'label': f'{k0}-o',
                    'ls': 'None',
                    'marker': 'o',
                    'ms': 4,
                    'c': color,
                },
            }

        # unit vectors
        if 'v' in elements:

            vinr = np.hypot(cx + vinx, cy + viny) - cr
            v0r = np.hypot(cx + v0x, cy + v0y) - cr
            v1r = np.hypot(cx + v1x, cy + v1y) - cr

            # dict

            dplot[k0]['v-nin'] = {
                'x': cx,
                'y': cy,
                'z': cz,
                'r': cr,
                'ux': vinx,
                'uy': viny,
                'uz': vinz,
                'ur': vinr,
                'props': {
                    'label': f'{k0}-nin',
                    'fc': 'r',
                    'color': 'r',
                },
            }

            dplot[k0]['v-e0'] = {
                'x': cx,
                'y': cy,
                'z': cz,
                'r': cr,
                'ux': v0x,
                'uy': v0y,
                'uz': v0z,
                'ur': v0r,
                'props': {
                    'label': f'{k0}-e0',
                    'fc': 'g',
                    'color': 'g',
                },
            }

            dplot[k0]['v-e1'] = {
                'x': cx,
                'y': cy,
                'z': cz,
                'r': cr,
                'ux': v1x,
                'uy': v1y,
                'uz': v1z,
                'ur': v1r,
                'props': {
                    'label': f'{k0}-e1',
                    'fc': 'b',
                    'color': 'b',
                },
            }

        # rowland / axis for curved optics
        if 'r' in elements and cls in ['crystal', 'grating']:

            if v0['type'] not in ['cylindrical', 'spherical', 'toroidal']:
                continue

            theta = np.linspace(-1, 1, 50) * np.pi
            if v0['type'] == 'cylindrical':
                c2x = cx + ninx * rc
                c2y = cy + niny * rc
                c2z = cz + ninz * rc
                px = c2x + np.r_[-1, 1] * axis_length * eax[0]
                py = c2y + np.r_[-1, 1] * axis_length * eax[1]
                pz = c2z + np.r_[-1, 1] * axis_length * eax[2]

                lab = f'{k0}-axis',

            elif v0['type'] == 'spherical':
                c2x = cx + ninx * 0.5 * rc
                c2y = cy + niny * 0.5 * rc
                c2z = cz + ninz * 0.5 * rc
                px = (
                    c2x
                    + 0.5 * rc * np.cos(theta) * (-ninx)
                    + 0.5 * rc * np.sin(theta) * e0x
                )
                py = (
                    c2y
                    + 0.5 * rc * np.cos(theta) * (-niny)
                    + 0.5 * rc * np.sin(theta) * e0y
                )
                pz = (
                    c2z
                    + 0.5 * rc * np.cos(theta) * (-ninz)
                    + 0.5 * rc * np.sin(theta) * e0z
                )

                lab = f'{k0}-rowland',

            elif v0['type'] == 'toroidal':
                c2x = cx + ninx * (rmin + rmax)
                c2y = cy + niny * (rmin + rmax)
                c2z = cz + ninz * (rmin + rmax)
                px = (
                    c2x
                    + rmax * np.cos(theta) * (-ninx)
                    + rmax * np.sin(theta) * emax[0]
                )
                py = (
                    c2y
                    + rmax * np.cos(theta) * (-niny)
                    + rmax * np.sin(theta) * emax[1]
                )
                pz = (
                    c2z
                    + rmax * np.cos(theta) * (-ninz)
                    + rmax * np.sin(theta) * emax[2]
                )

                lab = f'{k0}-majorR'

            dplot[k0]['r'] = {
                'x': px,
                'y': py,
                'z': pz,
                'r': np.hypot(px, py),
                'props': {
                    'label': lab,
                    'ls': '--',
                    'lw': 1.,
                    'color': color,
                },
            }

    return dplot


# ##################################################################
# ##################################################################
#                   Wavelength from angle
# ##################################################################


def get_lamb_from_angle(
    coll=None,
    key=None,
    key_cam=None,
    lamb=None,
    rocking_curve=None,
):
    """"""

    # ----------
    # check

    # key
    lok = list(coll.dobj.get('diagnostic', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # key_cam
    lok = list(coll.dobj['diagnostic'][key]['doptics'].keys())
    key_cam = ds._generic_check._check_var(
        key_cam, 'key_cam',
        types=str,
        allowed=lok,
    )

    # doptics
    doptics = coll.dobj['diagnostic'][key]['doptics'][key_cam]
    if 'crystal' not in doptics['cls']:
        raise Exception(f"Diag '{key}' is not a spectro!")

    kcryst = doptics['optics'][doptics['cls'].index('crystal')]

    dok = {
        'lamb': 'alpha',
        'lambmin': 'amin',
        'lambmax': 'amax',
        'res': 'res',
    }
    lok = list(dok.keys())
    lamb = ds._generic_check._check_var(
        lamb, 'lamb',
        types=str,
        allowed=lok,
    )

    # ----------
    # compute

    lv = []
    lk = ['lamb', 'lambmin', 'lambmax']
    for kk in lk:
        if lamb in [kk, 'res']:

            if kk == 'lamb':
                klos = coll.dobj['diagnostic'][key]['doptics'][key_cam]['los']
                ka = coll.dobj['rays'][klos][dok[kk]]
                ang = coll.ddata[ka]['data'][0, ...]
                ref = coll.ddata[ka]['ref'][1:]
            else:
                ka = coll.dobj['diagnostic'][key]['doptics'][key_cam][dok[kk]]
                ang = coll.ddata[ka]['data']
                ref = coll.ddata[ka]['ref']

            dd = coll.get_crystal_bragglamb(
                key=kcryst,
                bragg=ang,
                rocking_curve=rocking_curve,
            )[1]
            if lamb == kk:
                data = dd
            else:
                lv.append(dd)

    if lamb == 'res':
        data = lv[0] / (lv[2] - lv[1])

    return data, ref


# ##################################################################
# ##################################################################
#                   get data
# ##################################################################


def _get_data(
    coll=None,
    key=None,
    key_cam=None,
    data=None,
    rocking_curve=None,
    **kwdargs,
    ):

    # key, key_cam
    key, key_cam = coll.get_diagnostic_cam(key=key, key_cam=key_cam)
    spectro = coll.dobj['diagnostic'][key]['spectro']
    # is2d = coll.dobj['diagnostic'][key]['is2d']

    # basic check on data
    if data is not None:
        lquant = ['etendue', 'amin', 'amax']  # 'los'
        lcomp = ['length', 'tangency radius', 'alpha']
        if spectro:
            lcomp += ['lamb', 'lambmin', 'lambmax', 'res']

        data = ds._generic_check._check_var(
            data, 'data',
            types=str,
            allowed=lquant + lcomp,
        )

    # build ddata
    ddata = {}
    # comp = False
    if data is None or data in lquant:

        # --------------------------
        # data is None => kwdargs

        if data is None:
            # check kwdargs
            dparam = coll.get_param(which='data', returnas=dict)
            lkout = [k0 for k0 in kwdargs.keys() if k0 not in dparam.keys()]

            if len(lkout) > 0:
                msg= (
                    "The following args correspond to no data parameter:\n"
                    + "\n".join([f"\t- {k0}" for k0 in lkout])
                )
                raise Exception(msg)

            # list all available data
            lok = [
                k0 for k0, v0 in coll.ddata.items()
                if v0.get('camera') in key_cam
            ]

            # Adjust with kwdargs
            if len(kwdargs) > 0:
                lok2 = coll.select(
                    which='data', log='all', returnas=str, **kwdargs,
                )
                lok = [k0 for k0 in lok2 if k0 in lok]

            # check there is 1 data per cam
            lcam = [
                coll.ddata[k0]['camera'] for k0 in lok
                if coll.ddata[k0]['camera'] in key_cam
            ]

            if len(set(lcam)) > len(key_cam):
                msg = (
                    "There are more / less data identified than cameras:\n"
                    f"\t- key_cam:  {key_cam}\n"
                    f"\t- data cam: {lcam}\n"
                    f"\t- data: {data}"
                )
                raise Exception(msg)
            elif len(set(lcam)) < len(key_cam):
                pass

            # reorder
            ddata = {
                cc: lok[lcam.index(cc)]
                for cc in key_cam if cc in lcam
            }

        # -----------------
        # data in lquant

        elif data in lquant:
            for cc in key_cam:
                # if data == 'los':
                #     kr = coll.dobj['diagnostic'][key]['doptics'][cc][data]
                #     dd = coll.dobj['rays'][kr]['pts']
                # else:
                dd = coll.dobj['diagnostic'][key]['doptics'][cc][data]
                lc = [
                    isinstance(dd, str) and dd in coll.ddata.keys(),
                    # isinstance(dd, tuple)
                    # and all([isinstance(di, str) for di in dd])
                    # and all([di in coll.ddata.keys() for di in dd])
                ]
                if lc[0]:
                    ddata[cc] = dd
                # elif lc[1]:
                #     ddata[cc] = list(dd)
                elif dd is None:
                    pass
                else:
                    msg = f"Unknown data: '{data}'"
                    raise Exception(msg)

        # dref
        dref = {
            k0: coll.ddata[v0]['ref']
            for k0, v0 in ddata.items()
        }

        # get actual data
        ddata = {
            k0 : coll.ddata[v0]['data']
            for k0, v0 in ddata.items()
        }

    # --------------------
    # data to be computed

    elif data in lcomp:

        # comp = True
        ddata = {}
        dref = {}

        if data in ['lamb', 'lambmin', 'lambmax', 'res']:
            for cc in key_cam:
               ddata[cc], dref[cc] = coll.get_diagnostic_lamb(
                   key=key,
                   key_cam=cc,
                   rocking_curve=rocking_curve,
                   lamb=data,
               )

        elif data in ['length', 'tangency radius', 'alpha']:
            for cc in key_cam:
                ddata[cc], _, dref[cc] = coll.get_rays_quantity(
                    key=key,
                    key_cam=cc,
                    quantity=data,
                    segment=-1,
                    lim_to_segments=False,
                )

    return ddata, dref


# ##################################################################
# ##################################################################
#                   concatenate data
# ##################################################################


# def _concatenate_check(
    # coll=None,
    # key=None,
    # key_cam=None,
    # data=None,
    # rocking_curve=None,
    # returnas=None,
    # # naming
    # key_data=None,
    # key_ref=None,
    # **kwdargs,
    # ):

    # # -------------
    # # key, key_cam
    # # -------------

    # key, key_cam = coll.get_diagnostic_cam(key=key, key_cam=key_cam)
    # spectro = coll.dobj['diagnostic'][key]['spectro']
    # is2d = coll.dobj['diagnostic'][key]['is2d']
    # # stack = coll.dobj['diagnostic'][key]['stack']

    # if is2d and len(key_cam) > 1:
        # msg = (
            # "Cannot yet concatenate several 2d cameras\n"
            # "\t- key: '{key}'\n"
            # "\t- is2d: {is2d}\n"
            # "\t- key_cam: {key_cam}\n"
        # )
        # raise NotImplementedError(msg)

    # # ---------------
    # # build ddata
    # # -------------

    # # basic check on data
    # if data is not None:
        # lquant = ['los', 'etendue', 'amin', 'amax']
        # lcomp = ['tangency radius']
        # if spectro:
            # lcomp += ['lamb', 'lambmin', 'lambmax', 'res']

        # data = ds._generic_check._check_var(
            # data, 'data',
            # types=str,
            # allowed=lquant + lcomp,
        # )

    # # build ddata
    # ddata = {}
    # comp = False
    # if data is None or data in lquant:

        # # --------------------------
        # # data is None => kwdargs

        # if data is None:
            # # check kwdargs
            # dparam = coll.get_param(which='data', returnas=dict)
            # lkout = [k0 for k0 in kwdargs.keys() if k0 not in dparam.keys()]

            # if len(lkout) > 0:
                # msg= (
                    # "The following args correspond to no data parameter:\n"
                    # + "\n".join([f"\t- {k0}" for k0 in lkout])
                # )
                # raise Exception(msg)

            # # list all available data
            # lok = [
                # k0 for k0, v0 in coll.ddata.items()
                # if v0.get('camera') in key_cam
            # ]

            # # Adjust with kwdargs
            # if len(kwdargs) > 0:
                # lok2 = coll.select(
                    # which='data', log='all', returnas=str, **kwdargs,
                # )
                # lok = [k0 for k0 in lok2 if k0 in lok]

            # # check there is 1 data per cam
            # lcam = [
                # coll.ddata[k0]['camera'] for k0 in lok
                # if coll.ddata[k0]['camera'] in key_cam
            # ]

            # if len(set(lcam)) > len(key_cam):
                # msg = (
                    # "There are more / less data identified than cameras:\n"
                    # f"\t- key_cam:  {key_cam}\n"
                    # f"\t- data cam: {lcam}\n"
                    # f"\t- data: {data}"
                # )
                # raise Exception(msg)
            # elif len(set(lcam)) < len(key_cam):
                # pass

            # # reorder
            # ddata = {
                # cc: [lok[lcam.index(cc)]]
                # for cc in key_cam if cc in lcam
            # }

        # # -----------------
        # # data in lquant

        # elif data in lquant:
            # for cc in key_cam:
                # if data == 'los':
                    # kr = coll.dobj['diagnostic'][key]['doptics'][cc][data]
                    # dd = coll.dobj['rays'][kr]['pts']
                # else:
                    # dd = coll.dobj['diagnostic'][key]['doptics'][cc][data]
                # lc = [
                    # isinstance(dd, str) and dd in coll.ddata.keys(),
                    # isinstance(dd, tuple)
                    # and all([isinstance(di, str) for di in dd])
                    # and all([di in coll.ddata.keys() for di in dd])
                # ]
                # if lc[0]:
                    # ddata[cc] = [dd]
                # elif lc[1]:
                    # ddata[cc] = list(dd)
                # elif dd is None:
                    # pass
                # else:
                    # msg = f"Unknown data: '{data}'"
                    # raise Exception(msg)

        # # dref
        # dref = {
            # k0: [coll.ddata[k1]['ref'] for k1 in v0]
            # for k0, v0 in ddata.items()
        # }

    # # --------------------
    # # data to be computed

    # # TBF
    # elif data in lcomp:

        # comp = True
        # ddata = {[None] for cc in key_cam}
        # dref = {[None] for cc in key_cam}

        # if data in ['lamb', 'lambmin', 'lambmax', 'res']:
            # for cc in key_cam:
               # ddata[cc][0], dref[cc][0] = coll.get_diagnostic_lamb(
                   # key=key,
                   # key_cam=cc,
                   # rocking_curve=rocking_curve,
                   # lamb=data,
               # )

        # elif data == 'tangency radius':
            # ddata[cc][0], _, dref[cc][0] = coll.get_rays_tangency_radius(
                # key=key,
                # key_cam=key_cam,
                # segment=-1,
                # lim_to_segments=False,
            # )

    # # -----------------------------------
    # # Final safety checks and adjustments
    # # -----------------------------------

    # # adjust key_cam
    # key_cam = [cc for cc in key_cam if cc in ddata.keys()]

    # # ddata vs dref vs key_cam
    # lcd = sorted(list(ddata.keys()))
    # lcr = sorted(list(dref.keys()))
    # if not (sorted(key_cam) == lcd == lcr):
        # msg = (
            # "Wrong keys!\n"
            # f"\t- key_cam: {key_cam}\n"
            # f"\t- ddata.keys(): {lcd}\n"
            # f"\t- dref.keys(): {lcr}\n"
        # )
        # raise Exception(msg)

    # # nb of data per cam
    # ln = [len(v0) for v0 in ddata.values()]
    # if len(set(ln)) != 1:
        # msg = (
            # "Not the same number of data per cameras!\n"
            # + str(ddata)
        # )
        # raise Exception(msg)

    # # check shapes and ndim
    # dshapes = {
        # k0: [tuple([coll.dref[k2]['size'] for k2 in k1]) for k1 in v0]
        # for k0, v0 in dref.items()
    # }

    # # all same ndim
    # ndimref = None
    # for k0, v0 in dshapes.items():
        # lndim = [len(v1) for v1 in v0]
        # if len(set(lndim)) > 1:
            # msg = "All data must have same number of dimensions!\n{dshapes}"
            # raise Exception(msg)
        # if ndimref is None:
            # ndimref = lndim[0]
        # elif lndim[0] != ndimref:
            # msg = "All data must have same number of dimensions!\n{dshapes}"
            # raise Exception(msg)

    # # check indices of camera ref in data ref
    # indref = None
    # for k0, v0 in dref.items():
        # for v1 in v0:
            # ind = [v1.index(rr) for rr in coll.dobj['camera'][k0]['dgeom']['ref']]
            # if indref is None:
                # indref = ind
            # elif ind != indref:
                # msg = "All data must have same index of cam ref!\n{drf}"
                # raise Exception(msg)

    # if len(indref) > 1:
        # msg = "Cannot conatenate 2d cameras so far"
        # raise Exception(msg)

    # # check all shapes other than camera shapes are identical
    # if ndimref > len(indref):
        # ind = np.delete(np.arange(0, ndimref), indref)
        # shape0 = tuple(np.r_[dshapes[key_cam[0]][0]][ind])
        # lcout = [
            # cc for cc in key_cam
            # if any([tuple(np.r_[vv][ind]) != shape0 for vv in dshapes[cc]])
        # ]
        # if len(lcout) > 0:
            # msg = (
                # "The cameras data shall all have same shape (except pixels)\n"
                # + str(dshapes)
            # )
            # raise Exception(msg)

    # # check indices of camera ref in data ref
    # ref = None
    # for k0, v0 in dref.items():
        # for v1 in v0:
            # if ref is None:
                # ref = [
                    # None if ii == indref[0] else rr
                    # for ii, rr in enumerate(v1)
                # ]
            # else:
                # lc = [
                    # v1[ii] == ref[ii] for ii in range(ndimref)
                    # if ii not in indref
                # ]
                # if not all(lc):
                    # msg = (
                        # "All ref axcept the camera ref must be the same!\n"
                        # f"\t- ref: {ref}\n"
                        # f"\t- indref: {indref}\n"
                        # f"\t- ndimref: {ndimref}\n"
                        # f"\t- v1: {v1}\n"
                        # f"\t- lc: {lc}\n"
                        # + str(dref)
                    # )
                    # raise Exception(msg)

    # # -----------------------------------
    # # keys for new data and ref
    # # -----------------------------------

    # if key_data is None:
        # if data in lquant + lcomp:
            # if data == 'los':
                # key_data = [
                    # f'{key}_los_ptsx',
                    # f'{key}_los_ptsy',
                    # f'{key}_los_ptsz',
                # ]
            # else:
                # key_data = [f'{key}_{data}']
        # else:
            # key_data = [f'{key}_data']
    # elif isinstance(key_data, str):
        # key_data = [key_data]

    # if key_ref is None:
        # key_ref = f'{key}_npix'

    # ref = tuple([key_ref if rr is None else rr for rr in ref])

    # # -----------------------------------
    # # Other variables
    # # -----------------------------------

    # # returnas
    # returnas = ds._generic_check._check_var(
        # returnas, 'returnas',
        # default='Datastock',
        # allowed=[dict, 'Datastock'],
    # )

    # return (
        # key, key_cam, is2d,
        # ddata, ref, comp,
        # dshapes, ndimref, indref,
        # key_data, key_ref,
        # returnas,
    # )


def _concatenate_data_check(
    coll=None,
    key=None,
    key_data=None,
    flat=None,
):

    # ------------
    # key

    key, key_cam = coll.get_diagnostic_cam(key=key, key_cam=None)
    spectro = coll.dobj['diagnostic'][key]['spectro']
    is2d = coll.dobj['diagnostic'][key]['is2d']
    stack = coll.dobj['diagnostic'][key]['stack']

    # ------------
    # key_data

    # key_data
    if isinstance(key_data, str):

        if key_data in coll.ddata.keys():
            key_data = [key_data]

        else:
            lok = list(coll.dobj['diagnostic'][key].get('dsignal').keys())
            key_data = ds._generic_check._check_var(
                key_data, 'key_data',
                types=str,
                allowed=lok,
            )

            key_data = coll.dobj['diagnostic'][key]['dsignal'][key_data]['data']

    # basic check
    c0 = (
        isinstance(key_data, list)
        and all([
            isinstance(kk, str) and kk in coll.ddata.keys()
            for kk in key_data
        ])
    )
    if not c0:
        msg = (
            "Arg key_data must be a list of valid data keys!\n"
            f"Provided: {key_data}"
        )
        raise Exception(msg)

    # check cameras
    dout = {
        k0: coll.ddata[k0].get('camera')
        for k0 in key_data
        if coll.ddata[k0].get('camera') not in key_cam
    }
    if len(dout) > 0:
        lstr = [f"\t- {k0}: {v0}" for k0, v0 in dout.items()]
        msg = (
            f"The following data refers to no known camera in diag '{key}:\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # check unicity of cameras
    lcam = [coll.ddata[k0]['camera'] for k0 in key_data]
    if len(set(lcam)) > len(lcam):
        msg = (
            f"Non-unique camera references for diag '{key}'\n:"
            f"\t- key_data = {key_data}\n"
            f"\t- lcam     = {lcam}\n"
        )
        raise Exception(msg)

    # re-order
    lk = []
    for k0 in coll.dobj['diagnostic'][key]['camera']:
        if k0 in lcam:
            lk.append(key_data[lcam.index(k0)])

    key_data = lk
    lcam = [coll.ddata[k0]['camera'] for k0 in key_data]

    # ref uniformity
    lref = [coll.ddata[k0]['ref'] for k0 in key_data]
    if any([len(ref) != len(lref[0]) for ref in lref[1:]]):
        msg = (
            f"Non uniform refs for data in diag '{key}':\n"
            f"\t- key_data = {key_data}\n"
            f"\t- lref = {lref}\n"
        )
        raise Exception(msg)

    # ref axis
    laxcam = [
        [
            ref.index(rr)
            for rr in coll.dobj['camera'][lcam[ii]]['dgeom']['ref']
        ]
        for ii, ref in enumerate(lref)
    ]
    if any([len(ax) != len(laxcam[0]) for ax in laxcam[1:]]):
        msg = (
            f"Non-uniform camera concatenation axis for diag '{key}':\n"
            f"\t- key_data: {key_data}\n"
            f"\t- laxcam:   {laxcam}"
        )
        import pdb; pdb.set_trace()     # DB
        raise Exception(msg)

    laxcam = np.array(laxcam)
    if not np.allclose(laxcam, laxcam[0:1, :]):
        msg = (
            f"Non-uniform axis for concatenation for diag '{key}':\n"
            f"\t- laxcam: {laxcam}"
        )
        raise Exception(msg)
    axcam = laxcam[0]

    if len(axcam) != 1 + is2d:
        msg = (
            f"ref not consistent with is2d for diag '{key}':\n"
            f"\t- is2d: {is2d}\n"
            f"\t- axcam: {axcam}\n"
        )
        raise Exception(msg)

    ref = list(lref[0])
    for ii in axcam:
        ref[ii] = None

    # ------
    # flat

    flat = ds._generic_check._check_var(
        flat, 'flat',
        types=bool,
        default=is2d,
    )

    return key, key_data, is2d, stack, ref, flat


def _concatenate_data(
    coll=None,
    key=None,
    key_data=None,
    flat=None,
):

    # ------------
    # check inputs

    key, key_data, is2d, stack, ref, flat = _concatenate_data_check(
        coll=coll,
        key=key,
        key_data=key_data,
        flat=flat,
    )

    # ------------
    # prepare

    if is2d:
        ax0 = ref.index(None)
        ax1 = len(ref) - 1 - ref[::-1].index(None)
        if flat:
            ldata = []
            for k0 in key_data:
                sh = list(coll.ddata[k0]['data'].shape)
                size = sh[ax0] * sh[ax1]
                sh = tuple(np.r_[sh[:ax0], size, sh[ax1+1:]].astype(int))
                ldata.append(coll.ddata[k0]['data'].reshape(sh))
            axis = ax0
        else:
            axis = ax0 if stack == 'horizontal' else ax1
            ldata = [coll.ddata[k0]['data'] for k0 in key_data]
            lsize = [dd.shape[axis] for dd in ldata]
            if len(set(lsize)) != 1:
                msg = (
                    "Data for diag '{key}' cannot be stacked {stack}:\n"
                    f"\t- shapes: {[dd.shape for dd in ldata]}\n"
                    f"\t- axis: {axis}"
                )
                raise Exception(msg)

    else:
        ldata = [coll.ddata[k0]['data'] for k0 in key_data]
        axis = ref.index(None)

    # ------------
    # concatenate

    data = np.concatenate(tuple(ldata), axis=axis)
    units = coll.ddata[key_data[0]]['units']

    # dind
    i0 = 0
    dind = {}
    for ii, k0 in enumerate(key_data):
        npix = ldata[ii].shape[axis]
        ind = i0 + np.arange(0, npix)
        dind[k0] = ind
        i0 += npix

    return {
        'data': data,
        'keys': key_data,
        'units': units,
        'ref': ref,
        'axis': axis,
        'flat': flat,
        'dind': dind,
    }



# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import datastock as ds


# specific
from . import _generic_check
from . import _generic_plot


# ##################################################################
# ##################################################################
#                           plot check
# ##################################################################


def _plot_diagnostic_check(
    coll=None,
    key=None,
    key_cam=None,
    # figure
    proj=None,
    data=None,
    rocking_curve=None,
    los_res=None,
    # interactivity
    color_dict=None,
    nlos=None,
    connect=None,
):

    # -------
    # key

    # key
    key, key_cam = coll.get_diagnostic_cam(key, key_cam)
    is2d = coll.dobj['diagnostic'][key]['is2d']
    spectro = coll.dobj['diagnostic'][key]['spectro']

    if spectro:
        assert len(key_cam) == 1
        doptics = coll.dobj['diagnostic'][key]['doptics'][key_cam[0]]
        kcryst = doptics['optics'][doptics['ispectro'][0]]
    else:
        kcryst = None

    # -------
    # data

    defdata = 'etendue'
    c0 = data is None and coll.dobj['diagnostic'][key].get(defdata) is not None
    if c0:
        data = defdata

    ddata, dref = coll.get_diagnostic_data(
        key=key,
        key_cam=key_cam,
        data=data,
    )

    ylab = None # f"{ddata[key_cam[0]]['quant']} ({ddata[key_cam[0]]['units']})"

    # -----
    # proj

    proj = _generic_plot._proj(
        proj=proj,
        pall=['cross', 'hor', '3d', 'camera'],
    )

    # ----------
    # los_res

    los_res = ds._generic_check._check_var(
        los_res, 'los_res',
        types=float,
        default=0.05,
        sign='> 0.',
    )

    # -------
    # color_dict

    color_dict = _check_color_dict(color_dict)

    # -------
    # nlos

    nlos = ds._generic_check._check_var(
        nlos, 'nlos',
        types=int,
        default=5,
    )

    # -------
    # connect

    connect = ds._generic_check._check_var(
        connect, 'connect',
        types=bool,
        default=True,
    )

    return (
        key,
        key_cam,
        is2d,
        proj,
        ddata,
        dref,
        los_res,
        color_dict,
        nlos,
        ylab,
        connect,
    )


def _check_color_dict(color_dict=None):
    if color_dict is None:
        lc = ['r', 'g', 'b', 'm', 'c', 'y']
        color_dict = {
            'x': lc,
            'y': lc,
        }
    return color_dict


# ##################################################################
# ##################################################################
#                           plot main
# ##################################################################


def _plot_diagnostic(
    coll=None,
    key=None,
    key_cam=None,
    optics=None,
    elements=None,
    proj=None,
    los_res=None,
    # data plot
    data=None,
    cmap=None,
    vmin=None,
    vmax=None,
    # config
    plot_config=None,
    # figure
    dax=None,
    dmargin=None,
    fs=None,
    wintit=None,
    # interactivity
    color_dict=None,
    nlos=None,
    dinc=None,
    connect=None,
):

    # ------------
    # check inputs

    (
        key,
        key_cam,
        is2d,
        proj,
        ddata,
        dref,
        los_res,
        color_dict,
        nlos,
        ylab,
        connect,
    ) = _plot_diagnostic_check(
        coll=coll,
        key=key,
        # figure
        proj=proj,
        data=data,
        los_res=los_res,
        # interactivity
        color_dict=color_dict,
        nlos=nlos,
        connect=connect,
    )

    # ------------
    # prepare data

    dplot = coll.get_diagnostic_dplot(
        key=key,
        key_cam=key_cam,
        optics=optics,
        elements=elements,
    )

    # -------------------------
    # prepare los interactivity

    # instanciate new Datastock
    coll2 = coll.__class__()

    # ---------------------
    # prepare los and ddata

    # dcamref
    dcamref, drefx, drefy = _prepare_dcamref(
        coll=coll,
        key_cam=key_cam,
        is2d=is2d,
    )

    # los
    dlos, dref_los = _prepare_los(
        coll=coll,
        coll2=coll2,
        dcamref=dcamref,
        key_diag=key,
        key_cam=key_cam,
        los_res=los_res,
    )

    # ddatax, ddatay
    reft, dkeyx, dkeyy, ddatax, ddatay, dextent = _prepare_datarefxy(
        coll=coll,
        coll2=coll2,
        dcamref=dcamref,
        drefx=drefx,
        drefy=drefy,
        ddata=ddata,
        is2d=is2d,
    )

    # -----------------
    # prepare figure

    if dax is None:

        dax = _generic_plot.get_dax_diag(
            proj=proj,
            dmargin=dmargin,
            fs=fs,
            wintit=wintit,
            tit=key,
            is2d=is2d,
            key_cam=key_cam,
        )

    dax = _generic_check._check_dax(dax=dax, main=proj[0])

    # -----------------
    # plot static parts

    for k0, v0 in dplot.items():

        for k1, v1 in v0.items():

            # cross
            kax = 'cross'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                if k1.startswith('v-'):
                    ax.quiver(
                        v1['r'],
                        v1['z'],
                        v1['ur'],
                        v1['uz'],
                        **v1.get('props', {}),
                    )

                else:
                    ax.plot(
                        v1['r'],
                        v1['z'],
                        **v1.get('props', {}),
                    )

            # hor
            kax = 'hor'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                if k1.startswith('v-'):
                    ax.quiver(
                        v1['x'],
                        v1['y'],
                        v1['ux'],
                        v1['uy'],
                        **v1.get('props', {}),
                    )

                else:
                    ax.plot(
                        v1['x'],
                        v1['y'],
                        **v1.get('props', {}),
                    )

            # 3d
            kax = '3d'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                if k1.startswith('v-'):
                    ax.quiver(
                        v1['x'],
                        v1['y'],
                        v1['z'],
                        v1['ux'],
                        v1['uy'],
                        v1['uz'],
                        **v1.get('props', {}),
                    )

                else:
                    ax.plot(
                        v1['x'],
                        v1['y'],
                        v1['z'],
                        **v1.get('props', {}),
                    )

            # plotting of 2d camera contour
            kax = k0
            if is2d and k0 in key_cam and dax.get(kax) is not None:
                ax = dax[kax]['handle']
                if k1 == 'o':
                    ax.plot(
                        v1['x0'],
                        v1['x1'],
                        **v1.get('props', {}),
                    )

    # plot data
    for k0 in key_cam:
        kax = k0
        if dax.get(kax) is not None and ddata is not None:
            ax = dax[kax]['handle']

            if is2d and reft is None:
                im = ax.imshow(
                    ddata[k0].T,
                    extent=dextent[k0],
                    cmap=cmap,
                    vmin=vmin,
                    vmax=vmax,
                    origin='lower',
                    interpolation='nearest',
                )
                plt.colorbar(im, ax=ax)

            elif reft is None:
                ax.plot(
                    ddata[k0],
                    c='k',
                    ls='-',
                    lw=1.,
                    marker='.',
                    ms=6,
                )
                ax.set_xlim(-1, ddata[k0].size)
                ax.set_ylabel(ylab)

                if vmin is not None:
                    ax.set_ylim(bottom=vmin)
                if vmax is not None:
                    ax.set_ylim(top=vmax)

    # ----------------
    # define and set dgroup

    if coll2 is not None:
        dgroup = {
            f'{k0}_x': {
                'ref': [drefx[k0]],
                'data': ['index'],
                'nmax': nlos,
            }
            for k0 in key_cam
        }

        if is2d:
            dgroup.update({
                'y': {
                    'ref': list(drefy.values()),
                    'data': ['index'],
                    'nmax': nlos,
                },
            })

        if reft is not None:
            dgroup.update({
                't': {
                    'ref': [reft],
                    'data': ['index'],
                    'nmax': 1,
                },
            })

    # ------------------
    # plot mobile parts

    for k0 in key_cam:

        if dlos[k0]['rays'] is not None:

            nan = np.full((dlos[k0]['x'].shape[0],), np.nan)

            # cross
            kax = 'cross'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                _add_camera_los_cross(
                    coll2=coll2,
                    k0=k0,
                    ax=ax,
                    kax=kax,
                    nlos=nlos,
                    dref_los=dref_los,
                    color_dict=color_dict,
                    nan=nan,
                )

            # hor
            kax = 'hor'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                for ii in range(nlos):
                    l0, = ax.plot(
                        nan,
                        nan,
                        c=color_dict['x'][ii],
                        ls='-',
                        lw=1.,
                    )

                    # add mobile
                    kl0 = f'{k0}-los-hor-{ii}'
                    coll2.add_mobile(
                        key=kl0,
                        handle=l0,
                        refs=dref_los[k0],
                        data=[f'{k0}_los_x', f'{k0}_los_y'],
                        dtype=['xdata', 'ydata'],
                        axes=kax,
                        ind=ii,
                    )

            # 3d
            kax = '3d'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                for ii in range(nlos):
                    l0, = ax.plot(
                        nan,
                        nan,
                        nan,
                        c=color_dict['x'][ii],
                        ls='-',
                        lw=1.,
                    )

                    # add mobile
                    kl0 = f'{k0}_los-3d-{ii}'
                    # coll2.add_mobile(
                    # key=kl0,
                    # handle=l0,
                    # refs=reflos,
                    # data=['index', 'index', 'index'],
                    # dtype=['xdata', 'ydata', 'zdata'],
                    # axes=kax,
                    # ind=ii,
                    # )

            # camera
            kax = k0
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                _add_camera_vlines_marker(
                    coll2=coll2,
                    dax=dax,
                    ax=ax,
                    kax=kax,
                    is2d=is2d,
                    k0=k0,
                    nlos=nlos,
                    ddatax=ddatax,
                    ddatay=ddatay,
                    drefx=drefx,
                    drefy=drefy,
                    dkeyx=dkeyx,
                    dkeyy=dkeyy,
                    color_dict=color_dict,
                )

    # -------
    # config

    if plot_config.__class__.__name__ == 'Config':

        kax = 'cross'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj=kax)

        kax = 'hor'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj=kax)

    # -------
    # connect

    if coll2.dobj.get('mobile') is not None:
        # add axes
        for ii, kax in enumerate(dax.keys()):
            harmonize = ii == len(dax.keys()) - 1
            coll2.add_axes(key=kax, harmonize=harmonize, **dax[kax])

        # connect
        if connect is True:
            coll2.setup_interactivity(
                kinter='inter0',
                dgroup=dgroup,
                dinc=dinc,
            )
            coll2.disconnect_old()
            coll2.connect()

            coll2.show_commands()
            return coll2
        else:
            return coll2, dgroup
    else:
        return dax


# ##################################################################
# ##################################################################
#                       Prepare
# ##################################################################


def _prepare_dcamref(
    coll=None,
    key_cam=None,
    is2d=None,
):
    dcamref = {
        k0: coll.dobj['camera'][k0]['dgeom']['ref']
        for k0 in key_cam
    }
    drefx = {k0: v0[0] for k0, v0 in dcamref.items()}

    if is2d:
        drefy = {k0: v0[1] for k0, v0 in dcamref.items()}
    else:
        drefy = None

    return dcamref, drefx, drefy


def _prepare_los(
    coll=None,
    coll2=None,
    dcamref=None,
    key_diag=None,
    key_cam=None,
    los_res=None,
):

    # create dlos
    dlos = {
        k0: {'rays': coll.dobj['diagnostic'][key_diag]['doptics'][k0]['los']}
        for k0 in key_cam
    }
    dref_los = {}

    # los on cams
    for k0, v0 in dcamref.items():
        for rr in v0:
            coll2.add_ref(key=rr, size=coll.dref[rr]['size'])

        # los
        if dlos[k0]['rays'] is not None:
            los_x, los_y, los_z = coll.sample_rays(
                key=dlos[k0]['rays'],
                res=los_res,
                mode='rel',
                concatenate=False,
            )
            los_r = np.hypot(los_x, los_y)
            reflos = coll.dobj['rays'][dlos[k0]['rays']]['ref']
            dref_los[k0] = (reflos[1:], reflos[1:])

            if reflos[0] not in coll2.dref.keys():
                coll2.add_ref(key=reflos[0], size=los_x.shape[0])

            coll2.add_data(key=f'{k0}_los_x', data=los_x, ref=reflos)
            coll2.add_data(key=f'{k0}_los_y', data=los_y, ref=reflos)
            coll2.add_data(key=f'{k0}_los_z', data=los_z, ref=reflos)
            coll2.add_data(key=f'{k0}_los_r', data=los_r, ref=reflos)

            # store x, y, z
            dlos[k0]['x'] = los_x
            dlos[k0]['y'] = los_y
            dlos[k0]['z'] = los_z

    return dlos, dref_los


def _prepare_datarefxy(
    coll=None,
    coll2=None,
    dcamref=None,
    drefx=None,
    drefy=None,
    ddata=None,
    is2d=None,
):
    # prepare dict
    dkeyx, ddatax = {}, {}
    if is2d:
        dkeyy, ddatay, dextent = {}, {}, {}
    else:
        dkeyy, ddatay, dextent = None, None, None

    # loop on cams
    for k0, v0 in dcamref.items():

        # datax, datay
        if ddata is not None:
            if is2d:
                dkeyx[k0], dkeyy[k0] = coll.dobj['camera'][k0]['dgeom']['cents']

                ddatax[k0] = coll.ddata[dkeyx[k0]]['data']
                ddatay[k0] = coll.ddata[dkeyy[k0]]['data']

                coll2.add_data(key=dkeyx[k0], data=ddatax[k0], ref=drefx[k0])
                coll2.add_data(key=dkeyy[k0], data=ddatay[k0], ref=drefy[k0])
            else:
                dkeyx[k0] = f'{k0}_i0'
                ddatax[k0] = np.arange(0, coll.dref[drefx[k0]]['size'])
                coll2.add_data(key=dkeyx[k0], data=ddatax[k0], ref=drefx[k0])

            # -------------------------
            # extent

            reft = None
            if is2d:
                if ddatax[k0].size == 1:
                    ddx = coll.ddata[coll.dobj['camera'][k0]['dgeom']['outline'][0]]['data']
                    ddx = np.max(ddx) - np.min(ddx)
                else:
                    ddx = ddatax[k0][1] - ddatax[k0][0]
                if ddatay[k0].size == 1:
                    ddy = coll.ddata[coll.dobj['camera'][k0]['dgeom']['outline'][1]]['data']
                    ddy = np.max(ddy) - np.min(ddy)
                else:
                    ddy = ddatay[k0][1] - ddatay[k0][0]

                dextent[k0] = (
                    ddatax[k0][0] - 0.5*ddx,
                    ddatax[k0][-1] + 0.5*ddx,
                    ddatay[k0][0] - 0.5*ddy,
                    ddatay[k0][-1] + 0.5*ddy,
                )

    return reft, dkeyx, dkeyy, ddatax, ddatay, dextent


# ##################################################################
# ##################################################################
#                       add mobile
# ##################################################################


def _add_camera_los_cross(
    coll2=None,
    k0=None,
    ax=None,
    kax=None,
    nlos=None,
    dref_los=None,
    color_dict=None,
    nan=None,
):

    for ii in range(nlos):
        l0, = ax.plot(
            nan,
            nan,
            c=color_dict['x'][ii],
            ls='-',
            lw=1.,
        )

        # add mobile
        kl0 = f'{k0}-los-cross-{ii}'
        coll2.add_mobile(
            key=kl0,
            handle=l0,
            refs=dref_los[k0],
            data=[f'{k0}_los_r', f'{k0}_los_z'],
            dtype=['xdata', 'ydata'],
            axes=kax,
            ind=ii,
        )


def _add_camera_vlines_marker(
    coll2=None,
    dax=None,
    ax=None,
    kax=None,
    is2d=None,
    k0=None,
    nlos=None,
    ddatax=None,
    ddatay=None,
    drefx=None,
    drefy=None,
    dkeyx=None,
    dkeyy=None,
    color_dict=None,
    suffix=None,
):

    if suffix is None:
        suffix = ''

    if is2d:
        for ii in range(nlos):
            mi, = ax.plot(
                ddatax[k0][0:1],
                ddatay[k0][0:1],
                marker='s',
                ms=6,
                markeredgecolor=color_dict['x'][ii],
                markerfacecolor='None',
            )

            km = f'{k0}_m{ii:02.0f}{suffix}'
            coll2.add_mobile(
                key=km,
                handle=mi,
                refs=[drefx[k0], drefy[k0]],
                data=[dkeyx[k0], dkeyy[k0]],
                dtype=['xdata', 'ydata'],
                axes=kax,
                ind=ii,
            )

        dax[kax].update(
            refx=[drefx[k0]],
            refy=[drefy[k0]],
            datax=[dkeyx[k0]],
            datay=[dkeyy[k0]],
        )

    else:

        for ii in range(nlos):
            lv = ax.axvline(
                ddatax[k0][0], c=color_dict['y'][ii], lw=1., ls='-',
            )
            kv = f'{k0}_v{ii:02.0f}{suffix}'
            coll2.add_mobile(
                key=kv,
                handle=lv,
                refs=drefx[k0],
                data=dkeyx[k0],
                dtype='xdata',
                axes=kax,
                ind=ii,
            )

        dax[kax].update(refx=[drefx[k0]], datax=[dkeyx[k0]])
